/*
 * Copyright (C) 2011 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class AllFields {
    static boolean sZ;
    static byte sB;
    static char sC;
    static double sD;
    static float sF;
    static int sI;
    static long sJ;
    static short sS;
    static Object sObject;
    static Object[] sObjectArray;

    boolean iZ;
    byte iB;
    char iC;
    double iD;
    float iF;
    int iI;
    long iJ;
    short iS;
    Object iObject;
    Object[] iObjectArray;
}
