/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class JNI {
  public static native boolean canDiscoverField(Class<?> klass, String name, boolean isStatic);
  public static native boolean canGetField(Class<?> klass, String name, boolean isStatic);
  public static native boolean canSetField(Class<?> klass, String name, boolean isStatic);

  public static native boolean canDiscoverMethod(Class<?> klass, String name, boolean isStatic);
  public static native boolean canInvokeMethodA(Class<?> klass, String name, boolean isStatic);
  public static native boolean canInvokeMethodV(Class<?> klass, String name, boolean isStatic);

  public static native boolean canDiscoverConstructor(Class<?> klass, String signature);
  public static native boolean canInvokeConstructorA(Class<?> klass, String signature);
  public static native boolean canInvokeConstructorV(Class<?> klass, String signature);
}
