/*
 * Copyright (C) 2008 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
    public static void main(String[] args) {
        Special special = new Special();
        special.callInner();
    }

    public static class Special {
        Blort mBlort = null;

        Special() {
            System.out.println("new Special()");
        }

        public void callInner() {
            mBlort.repaint();
        }
    }

    private class Blort {
        public void repaint() {
            System.out.println("shouldn't see this");
        }
    }

}
