/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

class ManyFieldsBase0 {
    public Object testField0000 = new Integer(0);
    public Object testField0001 = new Integer(1);
    public Object testField0002 = new Integer(2);
    public Object testField0003 = new Integer(3);
    public Object testField0004 = new Integer(4);
    public Object testField0005 = new Integer(5);
    public Object testField0006 = new Integer(6);
    public Object testField0007 = new Integer(7);
    public Object testField0008 = new Integer(8);
    public Object testField0009 = new Integer(9);
    public Object testField0010 = new Integer(10);
    public Object testField0011 = new Integer(11);
    public Object testField0012 = new Integer(12);
    public Object testField0013 = new Integer(13);
    public Object testField0014 = new Integer(14);
    public Object testField0015 = new Integer(15);
    public Object testField0016 = new Integer(16);
    public Object testField0017 = new Integer(17);
    public Object testField0018 = new Integer(18);
    public Object testField0019 = new Integer(19);
    public Object testField0020 = new Integer(20);
    public Object testField0021 = new Integer(21);
    public Object testField0022 = new Integer(22);
    public Object testField0023 = new Integer(23);
    public Object testField0024 = new Integer(24);
    public Object testField0025 = new Integer(25);
    public Object testField0026 = new Integer(26);
    public Object testField0027 = new Integer(27);
    public Object testField0028 = new Integer(28);
    public Object testField0029 = new Integer(29);
    public Object testField0030 = new Integer(30);
    public Object testField0031 = new Integer(31);
    public Object testField0032 = new Integer(32);
    public Object testField0033 = new Integer(33);
    public Object testField0034 = new Integer(34);
    public Object testField0035 = new Integer(35);
    public Object testField0036 = new Integer(36);
    public Object testField0037 = new Integer(37);
    public Object testField0038 = new Integer(38);
    public Object testField0039 = new Integer(39);
    public Object testField0040 = new Integer(40);
    public Object testField0041 = new Integer(41);
    public Object testField0042 = new Integer(42);
    public Object testField0043 = new Integer(43);
    public Object testField0044 = new Integer(44);
    public Object testField0045 = new Integer(45);
    public Object testField0046 = new Integer(46);
    public Object testField0047 = new Integer(47);
    public Object testField0048 = new Integer(48);
    public Object testField0049 = new Integer(49);
    public Object testField0050 = new Integer(50);
    public Object testField0051 = new Integer(51);
    public Object testField0052 = new Integer(52);
    public Object testField0053 = new Integer(53);
    public Object testField0054 = new Integer(54);
    public Object testField0055 = new Integer(55);
    public Object testField0056 = new Integer(56);
    public Object testField0057 = new Integer(57);
    public Object testField0058 = new Integer(58);
    public Object testField0059 = new Integer(59);
    public Object testField0060 = new Integer(60);
    public Object testField0061 = new Integer(61);
    public Object testField0062 = new Integer(62);
    public Object testField0063 = new Integer(63);
    public Object testField0064 = new Integer(64);
    public Object testField0065 = new Integer(65);
    public Object testField0066 = new Integer(66);
    public Object testField0067 = new Integer(67);
    public Object testField0068 = new Integer(68);
    public Object testField0069 = new Integer(69);
    public Object testField0070 = new Integer(70);
    public Object testField0071 = new Integer(71);
    public Object testField0072 = new Integer(72);
    public Object testField0073 = new Integer(73);
    public Object testField0074 = new Integer(74);
    public Object testField0075 = new Integer(75);
    public Object testField0076 = new Integer(76);
    public Object testField0077 = new Integer(77);
    public Object testField0078 = new Integer(78);
    public Object testField0079 = new Integer(79);
    public Object testField0080 = new Integer(80);
    public Object testField0081 = new Integer(81);
    public Object testField0082 = new Integer(82);
    public Object testField0083 = new Integer(83);
    public Object testField0084 = new Integer(84);
    public Object testField0085 = new Integer(85);
    public Object testField0086 = new Integer(86);
    public Object testField0087 = new Integer(87);
    public Object testField0088 = new Integer(88);
    public Object testField0089 = new Integer(89);
    public Object testField0090 = new Integer(90);
    public Object testField0091 = new Integer(91);
    public Object testField0092 = new Integer(92);
    public Object testField0093 = new Integer(93);
    public Object testField0094 = new Integer(94);
    public Object testField0095 = new Integer(95);
    public Object testField0096 = new Integer(96);
    public Object testField0097 = new Integer(97);
    public Object testField0098 = new Integer(98);
    public Object testField0099 = new Integer(99);
    public Object testField0100 = new Integer(100);
    public Object testField0101 = new Integer(101);
    public Object testField0102 = new Integer(102);
    public Object testField0103 = new Integer(103);
    public Object testField0104 = new Integer(104);
    public Object testField0105 = new Integer(105);
    public Object testField0106 = new Integer(106);
    public Object testField0107 = new Integer(107);
    public Object testField0108 = new Integer(108);
    public Object testField0109 = new Integer(109);
    public Object testField0110 = new Integer(110);
    public Object testField0111 = new Integer(111);
    public Object testField0112 = new Integer(112);
    public Object testField0113 = new Integer(113);
    public Object testField0114 = new Integer(114);
    public Object testField0115 = new Integer(115);
    public Object testField0116 = new Integer(116);
    public Object testField0117 = new Integer(117);
    public Object testField0118 = new Integer(118);
    public Object testField0119 = new Integer(119);
    public Object testField0120 = new Integer(120);
    public Object testField0121 = new Integer(121);
    public Object testField0122 = new Integer(122);
    public Object testField0123 = new Integer(123);
    public Object testField0124 = new Integer(124);
    public Object testField0125 = new Integer(125);
    public Object testField0126 = new Integer(126);
    public Object testField0127 = new Integer(127);
    public Object testField0128 = new Integer(128);
    public Object testField0129 = new Integer(129);
    public Object testField0130 = new Integer(130);
    public Object testField0131 = new Integer(131);
    public Object testField0132 = new Integer(132);
    public Object testField0133 = new Integer(133);
    public Object testField0134 = new Integer(134);
    public Object testField0135 = new Integer(135);
    public Object testField0136 = new Integer(136);
    public Object testField0137 = new Integer(137);
    public Object testField0138 = new Integer(138);
    public Object testField0139 = new Integer(139);
    public Object testField0140 = new Integer(140);
    public Object testField0141 = new Integer(141);
    public Object testField0142 = new Integer(142);
    public Object testField0143 = new Integer(143);
    public Object testField0144 = new Integer(144);
    public Object testField0145 = new Integer(145);
    public Object testField0146 = new Integer(146);
    public Object testField0147 = new Integer(147);
    public Object testField0148 = new Integer(148);
    public Object testField0149 = new Integer(149);
    public Object testField0150 = new Integer(150);
    public Object testField0151 = new Integer(151);
    public Object testField0152 = new Integer(152);
    public Object testField0153 = new Integer(153);
    public Object testField0154 = new Integer(154);
    public Object testField0155 = new Integer(155);
    public Object testField0156 = new Integer(156);
    public Object testField0157 = new Integer(157);
    public Object testField0158 = new Integer(158);
    public Object testField0159 = new Integer(159);
    public Object testField0160 = new Integer(160);
    public Object testField0161 = new Integer(161);
    public Object testField0162 = new Integer(162);
    public Object testField0163 = new Integer(163);
    public Object testField0164 = new Integer(164);
    public Object testField0165 = new Integer(165);
    public Object testField0166 = new Integer(166);
    public Object testField0167 = new Integer(167);
    public Object testField0168 = new Integer(168);
    public Object testField0169 = new Integer(169);
    public Object testField0170 = new Integer(170);
    public Object testField0171 = new Integer(171);
    public Object testField0172 = new Integer(172);
    public Object testField0173 = new Integer(173);
    public Object testField0174 = new Integer(174);
    public Object testField0175 = new Integer(175);
    public Object testField0176 = new Integer(176);
    public Object testField0177 = new Integer(177);
    public Object testField0178 = new Integer(178);
    public Object testField0179 = new Integer(179);
    public Object testField0180 = new Integer(180);
    public Object testField0181 = new Integer(181);
    public Object testField0182 = new Integer(182);
    public Object testField0183 = new Integer(183);
    public Object testField0184 = new Integer(184);
    public Object testField0185 = new Integer(185);
    public Object testField0186 = new Integer(186);
    public Object testField0187 = new Integer(187);
    public Object testField0188 = new Integer(188);
    public Object testField0189 = new Integer(189);
    public Object testField0190 = new Integer(190);
    public Object testField0191 = new Integer(191);
    public Object testField0192 = new Integer(192);
    public Object testField0193 = new Integer(193);
    public Object testField0194 = new Integer(194);
    public Object testField0195 = new Integer(195);
    public Object testField0196 = new Integer(196);
    public Object testField0197 = new Integer(197);
    public Object testField0198 = new Integer(198);
    public Object testField0199 = new Integer(199);
    public Object testField0200 = new Integer(200);
    public Object testField0201 = new Integer(201);
    public Object testField0202 = new Integer(202);
    public Object testField0203 = new Integer(203);
    public Object testField0204 = new Integer(204);
    public Object testField0205 = new Integer(205);
    public Object testField0206 = new Integer(206);
    public Object testField0207 = new Integer(207);
    public Object testField0208 = new Integer(208);
    public Object testField0209 = new Integer(209);
    public Object testField0210 = new Integer(210);
    public Object testField0211 = new Integer(211);
    public Object testField0212 = new Integer(212);
    public Object testField0213 = new Integer(213);
    public Object testField0214 = new Integer(214);
    public Object testField0215 = new Integer(215);
    public Object testField0216 = new Integer(216);
    public Object testField0217 = new Integer(217);
    public Object testField0218 = new Integer(218);
    public Object testField0219 = new Integer(219);
    public Object testField0220 = new Integer(220);
    public Object testField0221 = new Integer(221);
    public Object testField0222 = new Integer(222);
    public Object testField0223 = new Integer(223);
    public Object testField0224 = new Integer(224);
    public Object testField0225 = new Integer(225);
    public Object testField0226 = new Integer(226);
    public Object testField0227 = new Integer(227);
    public Object testField0228 = new Integer(228);
    public Object testField0229 = new Integer(229);
    public Object testField0230 = new Integer(230);
    public Object testField0231 = new Integer(231);
    public Object testField0232 = new Integer(232);
    public Object testField0233 = new Integer(233);
    public Object testField0234 = new Integer(234);
    public Object testField0235 = new Integer(235);
    public Object testField0236 = new Integer(236);
    public Object testField0237 = new Integer(237);
    public Object testField0238 = new Integer(238);
    public Object testField0239 = new Integer(239);
    public Object testField0240 = new Integer(240);
    public Object testField0241 = new Integer(241);
    public Object testField0242 = new Integer(242);
    public Object testField0243 = new Integer(243);
    public Object testField0244 = new Integer(244);
    public Object testField0245 = new Integer(245);
    public Object testField0246 = new Integer(246);
    public Object testField0247 = new Integer(247);
    public Object testField0248 = new Integer(248);
    public Object testField0249 = new Integer(249);
    public Object testField0250 = new Integer(250);
    public Object testField0251 = new Integer(251);
    public Object testField0252 = new Integer(252);
    public Object testField0253 = new Integer(253);
    public Object testField0254 = new Integer(254);
    public Object testField0255 = new Integer(255);
    public Object testField0256 = new Integer(256);
    public Object testField0257 = new Integer(257);
    public Object testField0258 = new Integer(258);
    public Object testField0259 = new Integer(259);
    public Object testField0260 = new Integer(260);
    public Object testField0261 = new Integer(261);
    public Object testField0262 = new Integer(262);
    public Object testField0263 = new Integer(263);
    public Object testField0264 = new Integer(264);
    public Object testField0265 = new Integer(265);
    public Object testField0266 = new Integer(266);
    public Object testField0267 = new Integer(267);
    public Object testField0268 = new Integer(268);
    public Object testField0269 = new Integer(269);
    public Object testField0270 = new Integer(270);
    public Object testField0271 = new Integer(271);
    public Object testField0272 = new Integer(272);
    public Object testField0273 = new Integer(273);
    public Object testField0274 = new Integer(274);
    public Object testField0275 = new Integer(275);
    public Object testField0276 = new Integer(276);
    public Object testField0277 = new Integer(277);
    public Object testField0278 = new Integer(278);
    public Object testField0279 = new Integer(279);
    public Object testField0280 = new Integer(280);
    public Object testField0281 = new Integer(281);
    public Object testField0282 = new Integer(282);
    public Object testField0283 = new Integer(283);
    public Object testField0284 = new Integer(284);
    public Object testField0285 = new Integer(285);
    public Object testField0286 = new Integer(286);
    public Object testField0287 = new Integer(287);
    public Object testField0288 = new Integer(288);
    public Object testField0289 = new Integer(289);
    public Object testField0290 = new Integer(290);
    public Object testField0291 = new Integer(291);
    public Object testField0292 = new Integer(292);
    public Object testField0293 = new Integer(293);
    public Object testField0294 = new Integer(294);
    public Object testField0295 = new Integer(295);
    public Object testField0296 = new Integer(296);
    public Object testField0297 = new Integer(297);
    public Object testField0298 = new Integer(298);
    public Object testField0299 = new Integer(299);
    public Object testField0300 = new Integer(300);
    public Object testField0301 = new Integer(301);
    public Object testField0302 = new Integer(302);
    public Object testField0303 = new Integer(303);
    public Object testField0304 = new Integer(304);
    public Object testField0305 = new Integer(305);
    public Object testField0306 = new Integer(306);
    public Object testField0307 = new Integer(307);
    public Object testField0308 = new Integer(308);
    public Object testField0309 = new Integer(309);
    public Object testField0310 = new Integer(310);
    public Object testField0311 = new Integer(311);
    public Object testField0312 = new Integer(312);
    public Object testField0313 = new Integer(313);
    public Object testField0314 = new Integer(314);
    public Object testField0315 = new Integer(315);
    public Object testField0316 = new Integer(316);
    public Object testField0317 = new Integer(317);
    public Object testField0318 = new Integer(318);
    public Object testField0319 = new Integer(319);
    public Object testField0320 = new Integer(320);
    public Object testField0321 = new Integer(321);
    public Object testField0322 = new Integer(322);
    public Object testField0323 = new Integer(323);
    public Object testField0324 = new Integer(324);
    public Object testField0325 = new Integer(325);
    public Object testField0326 = new Integer(326);
    public Object testField0327 = new Integer(327);
    public Object testField0328 = new Integer(328);
    public Object testField0329 = new Integer(329);
    public Object testField0330 = new Integer(330);
    public Object testField0331 = new Integer(331);
    public Object testField0332 = new Integer(332);
    public Object testField0333 = new Integer(333);
    public Object testField0334 = new Integer(334);
    public Object testField0335 = new Integer(335);
    public Object testField0336 = new Integer(336);
    public Object testField0337 = new Integer(337);
    public Object testField0338 = new Integer(338);
    public Object testField0339 = new Integer(339);
    public Object testField0340 = new Integer(340);
    public Object testField0341 = new Integer(341);
    public Object testField0342 = new Integer(342);
    public Object testField0343 = new Integer(343);
    public Object testField0344 = new Integer(344);
    public Object testField0345 = new Integer(345);
    public Object testField0346 = new Integer(346);
    public Object testField0347 = new Integer(347);
    public Object testField0348 = new Integer(348);
    public Object testField0349 = new Integer(349);
    public Object testField0350 = new Integer(350);
    public Object testField0351 = new Integer(351);
    public Object testField0352 = new Integer(352);
    public Object testField0353 = new Integer(353);
    public Object testField0354 = new Integer(354);
    public Object testField0355 = new Integer(355);
    public Object testField0356 = new Integer(356);
    public Object testField0357 = new Integer(357);
    public Object testField0358 = new Integer(358);
    public Object testField0359 = new Integer(359);
    public Object testField0360 = new Integer(360);
    public Object testField0361 = new Integer(361);
    public Object testField0362 = new Integer(362);
    public Object testField0363 = new Integer(363);
    public Object testField0364 = new Integer(364);
    public Object testField0365 = new Integer(365);
    public Object testField0366 = new Integer(366);
    public Object testField0367 = new Integer(367);
    public Object testField0368 = new Integer(368);
    public Object testField0369 = new Integer(369);
    public Object testField0370 = new Integer(370);
    public Object testField0371 = new Integer(371);
    public Object testField0372 = new Integer(372);
    public Object testField0373 = new Integer(373);
    public Object testField0374 = new Integer(374);
    public Object testField0375 = new Integer(375);
    public Object testField0376 = new Integer(376);
    public Object testField0377 = new Integer(377);
    public Object testField0378 = new Integer(378);
    public Object testField0379 = new Integer(379);
    public Object testField0380 = new Integer(380);
    public Object testField0381 = new Integer(381);
    public Object testField0382 = new Integer(382);
    public Object testField0383 = new Integer(383);
    public Object testField0384 = new Integer(384);
    public Object testField0385 = new Integer(385);
    public Object testField0386 = new Integer(386);
    public Object testField0387 = new Integer(387);
    public Object testField0388 = new Integer(388);
    public Object testField0389 = new Integer(389);
    public Object testField0390 = new Integer(390);
    public Object testField0391 = new Integer(391);
    public Object testField0392 = new Integer(392);
    public Object testField0393 = new Integer(393);
    public Object testField0394 = new Integer(394);
    public Object testField0395 = new Integer(395);
    public Object testField0396 = new Integer(396);
    public Object testField0397 = new Integer(397);
    public Object testField0398 = new Integer(398);
    public Object testField0399 = new Integer(399);
    public Object testField0400 = new Integer(400);
    public Object testField0401 = new Integer(401);
    public Object testField0402 = new Integer(402);
    public Object testField0403 = new Integer(403);
    public Object testField0404 = new Integer(404);
    public Object testField0405 = new Integer(405);
    public Object testField0406 = new Integer(406);
    public Object testField0407 = new Integer(407);
    public Object testField0408 = new Integer(408);
    public Object testField0409 = new Integer(409);
    public Object testField0410 = new Integer(410);
    public Object testField0411 = new Integer(411);
    public Object testField0412 = new Integer(412);
    public Object testField0413 = new Integer(413);
    public Object testField0414 = new Integer(414);
    public Object testField0415 = new Integer(415);
    public Object testField0416 = new Integer(416);
    public Object testField0417 = new Integer(417);
    public Object testField0418 = new Integer(418);
    public Object testField0419 = new Integer(419);
    public Object testField0420 = new Integer(420);
    public Object testField0421 = new Integer(421);
    public Object testField0422 = new Integer(422);
    public Object testField0423 = new Integer(423);
    public Object testField0424 = new Integer(424);
    public Object testField0425 = new Integer(425);
    public Object testField0426 = new Integer(426);
    public Object testField0427 = new Integer(427);
    public Object testField0428 = new Integer(428);
    public Object testField0429 = new Integer(429);
    public Object testField0430 = new Integer(430);
    public Object testField0431 = new Integer(431);
    public Object testField0432 = new Integer(432);
    public Object testField0433 = new Integer(433);
    public Object testField0434 = new Integer(434);
    public Object testField0435 = new Integer(435);
    public Object testField0436 = new Integer(436);
    public Object testField0437 = new Integer(437);
    public Object testField0438 = new Integer(438);
    public Object testField0439 = new Integer(439);
    public Object testField0440 = new Integer(440);
    public Object testField0441 = new Integer(441);
    public Object testField0442 = new Integer(442);
    public Object testField0443 = new Integer(443);
    public Object testField0444 = new Integer(444);
    public Object testField0445 = new Integer(445);
    public Object testField0446 = new Integer(446);
    public Object testField0447 = new Integer(447);
    public Object testField0448 = new Integer(448);
    public Object testField0449 = new Integer(449);
    public Object testField0450 = new Integer(450);
    public Object testField0451 = new Integer(451);
    public Object testField0452 = new Integer(452);
    public Object testField0453 = new Integer(453);
    public Object testField0454 = new Integer(454);
    public Object testField0455 = new Integer(455);
    public Object testField0456 = new Integer(456);
    public Object testField0457 = new Integer(457);
    public Object testField0458 = new Integer(458);
    public Object testField0459 = new Integer(459);
    public Object testField0460 = new Integer(460);
    public Object testField0461 = new Integer(461);
    public Object testField0462 = new Integer(462);
    public Object testField0463 = new Integer(463);
    public Object testField0464 = new Integer(464);
    public Object testField0465 = new Integer(465);
    public Object testField0466 = new Integer(466);
    public Object testField0467 = new Integer(467);
    public Object testField0468 = new Integer(468);
    public Object testField0469 = new Integer(469);
    public Object testField0470 = new Integer(470);
    public Object testField0471 = new Integer(471);
    public Object testField0472 = new Integer(472);
    public Object testField0473 = new Integer(473);
    public Object testField0474 = new Integer(474);
    public Object testField0475 = new Integer(475);
    public Object testField0476 = new Integer(476);
    public Object testField0477 = new Integer(477);
    public Object testField0478 = new Integer(478);
    public Object testField0479 = new Integer(479);
    public Object testField0480 = new Integer(480);
    public Object testField0481 = new Integer(481);
    public Object testField0482 = new Integer(482);
    public Object testField0483 = new Integer(483);
    public Object testField0484 = new Integer(484);
    public Object testField0485 = new Integer(485);
    public Object testField0486 = new Integer(486);
    public Object testField0487 = new Integer(487);
    public Object testField0488 = new Integer(488);
    public Object testField0489 = new Integer(489);
    public Object testField0490 = new Integer(490);
    public Object testField0491 = new Integer(491);
    public Object testField0492 = new Integer(492);
    public Object testField0493 = new Integer(493);
    public Object testField0494 = new Integer(494);
    public Object testField0495 = new Integer(495);
    public Object testField0496 = new Integer(496);
    public Object testField0497 = new Integer(497);
    public Object testField0498 = new Integer(498);
    public Object testField0499 = new Integer(499);
    public Object testField0500 = new Integer(500);
    public Object testField0501 = new Integer(501);
    public Object testField0502 = new Integer(502);
    public Object testField0503 = new Integer(503);
    public Object testField0504 = new Integer(504);
    public Object testField0505 = new Integer(505);
    public Object testField0506 = new Integer(506);
    public Object testField0507 = new Integer(507);
    public Object testField0508 = new Integer(508);
    public Object testField0509 = new Integer(509);
    public Object testField0510 = new Integer(510);
    public Object testField0511 = new Integer(511);
    public Object testField0512 = new Integer(512);
    public Object testField0513 = new Integer(513);
    public Object testField0514 = new Integer(514);
    public Object testField0515 = new Integer(515);
    public Object testField0516 = new Integer(516);
    public Object testField0517 = new Integer(517);
    public Object testField0518 = new Integer(518);
    public Object testField0519 = new Integer(519);
    public Object testField0520 = new Integer(520);
    public Object testField0521 = new Integer(521);
    public Object testField0522 = new Integer(522);
    public Object testField0523 = new Integer(523);
    public Object testField0524 = new Integer(524);
    public Object testField0525 = new Integer(525);
    public Object testField0526 = new Integer(526);
    public Object testField0527 = new Integer(527);
    public Object testField0528 = new Integer(528);
    public Object testField0529 = new Integer(529);
    public Object testField0530 = new Integer(530);
    public Object testField0531 = new Integer(531);
    public Object testField0532 = new Integer(532);
    public Object testField0533 = new Integer(533);
    public Object testField0534 = new Integer(534);
    public Object testField0535 = new Integer(535);
    public Object testField0536 = new Integer(536);
    public Object testField0537 = new Integer(537);
    public Object testField0538 = new Integer(538);
    public Object testField0539 = new Integer(539);
    public Object testField0540 = new Integer(540);
    public Object testField0541 = new Integer(541);
    public Object testField0542 = new Integer(542);
    public Object testField0543 = new Integer(543);
    public Object testField0544 = new Integer(544);
    public Object testField0545 = new Integer(545);
    public Object testField0546 = new Integer(546);
    public Object testField0547 = new Integer(547);
    public Object testField0548 = new Integer(548);
    public Object testField0549 = new Integer(549);
    public Object testField0550 = new Integer(550);
    public Object testField0551 = new Integer(551);
    public Object testField0552 = new Integer(552);
    public Object testField0553 = new Integer(553);
    public Object testField0554 = new Integer(554);
    public Object testField0555 = new Integer(555);
    public Object testField0556 = new Integer(556);
    public Object testField0557 = new Integer(557);
    public Object testField0558 = new Integer(558);
    public Object testField0559 = new Integer(559);
    public Object testField0560 = new Integer(560);
    public Object testField0561 = new Integer(561);
    public Object testField0562 = new Integer(562);
    public Object testField0563 = new Integer(563);
    public Object testField0564 = new Integer(564);
    public Object testField0565 = new Integer(565);
    public Object testField0566 = new Integer(566);
    public Object testField0567 = new Integer(567);
    public Object testField0568 = new Integer(568);
    public Object testField0569 = new Integer(569);
    public Object testField0570 = new Integer(570);
    public Object testField0571 = new Integer(571);
    public Object testField0572 = new Integer(572);
    public Object testField0573 = new Integer(573);
    public Object testField0574 = new Integer(574);
    public Object testField0575 = new Integer(575);
    public Object testField0576 = new Integer(576);
    public Object testField0577 = new Integer(577);
    public Object testField0578 = new Integer(578);
    public Object testField0579 = new Integer(579);
    public Object testField0580 = new Integer(580);
    public Object testField0581 = new Integer(581);
    public Object testField0582 = new Integer(582);
    public Object testField0583 = new Integer(583);
    public Object testField0584 = new Integer(584);
    public Object testField0585 = new Integer(585);
    public Object testField0586 = new Integer(586);
    public Object testField0587 = new Integer(587);
    public Object testField0588 = new Integer(588);
    public Object testField0589 = new Integer(589);
    public Object testField0590 = new Integer(590);
    public Object testField0591 = new Integer(591);
    public Object testField0592 = new Integer(592);
    public Object testField0593 = new Integer(593);
    public Object testField0594 = new Integer(594);
    public Object testField0595 = new Integer(595);
    public Object testField0596 = new Integer(596);
    public Object testField0597 = new Integer(597);
    public Object testField0598 = new Integer(598);
    public Object testField0599 = new Integer(599);
    public Object testField0600 = new Integer(600);
    public Object testField0601 = new Integer(601);
    public Object testField0602 = new Integer(602);
    public Object testField0603 = new Integer(603);
    public Object testField0604 = new Integer(604);
    public Object testField0605 = new Integer(605);
    public Object testField0606 = new Integer(606);
    public Object testField0607 = new Integer(607);
    public Object testField0608 = new Integer(608);
    public Object testField0609 = new Integer(609);
    public Object testField0610 = new Integer(610);
    public Object testField0611 = new Integer(611);
    public Object testField0612 = new Integer(612);
    public Object testField0613 = new Integer(613);
    public Object testField0614 = new Integer(614);
    public Object testField0615 = new Integer(615);
    public Object testField0616 = new Integer(616);
    public Object testField0617 = new Integer(617);
    public Object testField0618 = new Integer(618);
    public Object testField0619 = new Integer(619);
    public Object testField0620 = new Integer(620);
    public Object testField0621 = new Integer(621);
    public Object testField0622 = new Integer(622);
    public Object testField0623 = new Integer(623);
    public Object testField0624 = new Integer(624);
    public Object testField0625 = new Integer(625);
    public Object testField0626 = new Integer(626);
    public Object testField0627 = new Integer(627);
    public Object testField0628 = new Integer(628);
    public Object testField0629 = new Integer(629);
    public Object testField0630 = new Integer(630);
    public Object testField0631 = new Integer(631);
    public Object testField0632 = new Integer(632);
    public Object testField0633 = new Integer(633);
    public Object testField0634 = new Integer(634);
    public Object testField0635 = new Integer(635);
    public Object testField0636 = new Integer(636);
    public Object testField0637 = new Integer(637);
    public Object testField0638 = new Integer(638);
    public Object testField0639 = new Integer(639);
    public Object testField0640 = new Integer(640);
    public Object testField0641 = new Integer(641);
    public Object testField0642 = new Integer(642);
    public Object testField0643 = new Integer(643);
    public Object testField0644 = new Integer(644);
    public Object testField0645 = new Integer(645);
    public Object testField0646 = new Integer(646);
    public Object testField0647 = new Integer(647);
    public Object testField0648 = new Integer(648);
    public Object testField0649 = new Integer(649);
    public Object testField0650 = new Integer(650);
    public Object testField0651 = new Integer(651);
    public Object testField0652 = new Integer(652);
    public Object testField0653 = new Integer(653);
    public Object testField0654 = new Integer(654);
    public Object testField0655 = new Integer(655);
    public Object testField0656 = new Integer(656);
    public Object testField0657 = new Integer(657);
    public Object testField0658 = new Integer(658);
    public Object testField0659 = new Integer(659);
    public Object testField0660 = new Integer(660);
    public Object testField0661 = new Integer(661);
    public Object testField0662 = new Integer(662);
    public Object testField0663 = new Integer(663);
    public Object testField0664 = new Integer(664);
    public Object testField0665 = new Integer(665);
    public Object testField0666 = new Integer(666);
    public Object testField0667 = new Integer(667);
    public Object testField0668 = new Integer(668);
    public Object testField0669 = new Integer(669);
    public Object testField0670 = new Integer(670);
    public Object testField0671 = new Integer(671);
    public Object testField0672 = new Integer(672);
    public Object testField0673 = new Integer(673);
    public Object testField0674 = new Integer(674);
    public Object testField0675 = new Integer(675);
    public Object testField0676 = new Integer(676);
    public Object testField0677 = new Integer(677);
    public Object testField0678 = new Integer(678);
    public Object testField0679 = new Integer(679);
    public Object testField0680 = new Integer(680);
    public Object testField0681 = new Integer(681);
    public Object testField0682 = new Integer(682);
    public Object testField0683 = new Integer(683);
    public Object testField0684 = new Integer(684);
    public Object testField0685 = new Integer(685);
    public Object testField0686 = new Integer(686);
    public Object testField0687 = new Integer(687);
    public Object testField0688 = new Integer(688);
    public Object testField0689 = new Integer(689);
    public Object testField0690 = new Integer(690);
    public Object testField0691 = new Integer(691);
    public Object testField0692 = new Integer(692);
    public Object testField0693 = new Integer(693);
    public Object testField0694 = new Integer(694);
    public Object testField0695 = new Integer(695);
    public Object testField0696 = new Integer(696);
    public Object testField0697 = new Integer(697);
    public Object testField0698 = new Integer(698);
    public Object testField0699 = new Integer(699);
    public Object testField0700 = new Integer(700);
    public Object testField0701 = new Integer(701);
    public Object testField0702 = new Integer(702);
    public Object testField0703 = new Integer(703);
    public Object testField0704 = new Integer(704);
    public Object testField0705 = new Integer(705);
    public Object testField0706 = new Integer(706);
    public Object testField0707 = new Integer(707);
    public Object testField0708 = new Integer(708);
    public Object testField0709 = new Integer(709);
    public Object testField0710 = new Integer(710);
    public Object testField0711 = new Integer(711);
    public Object testField0712 = new Integer(712);
    public Object testField0713 = new Integer(713);
    public Object testField0714 = new Integer(714);
    public Object testField0715 = new Integer(715);
    public Object testField0716 = new Integer(716);
    public Object testField0717 = new Integer(717);
    public Object testField0718 = new Integer(718);
    public Object testField0719 = new Integer(719);
    public Object testField0720 = new Integer(720);
    public Object testField0721 = new Integer(721);
    public Object testField0722 = new Integer(722);
    public Object testField0723 = new Integer(723);
    public Object testField0724 = new Integer(724);
    public Object testField0725 = new Integer(725);
    public Object testField0726 = new Integer(726);
    public Object testField0727 = new Integer(727);
    public Object testField0728 = new Integer(728);
    public Object testField0729 = new Integer(729);
    public Object testField0730 = new Integer(730);
    public Object testField0731 = new Integer(731);
    public Object testField0732 = new Integer(732);
    public Object testField0733 = new Integer(733);
    public Object testField0734 = new Integer(734);
    public Object testField0735 = new Integer(735);
    public Object testField0736 = new Integer(736);
    public Object testField0737 = new Integer(737);
    public Object testField0738 = new Integer(738);
    public Object testField0739 = new Integer(739);
    public Object testField0740 = new Integer(740);
    public Object testField0741 = new Integer(741);
    public Object testField0742 = new Integer(742);
    public Object testField0743 = new Integer(743);
    public Object testField0744 = new Integer(744);
    public Object testField0745 = new Integer(745);
    public Object testField0746 = new Integer(746);
    public Object testField0747 = new Integer(747);
    public Object testField0748 = new Integer(748);
    public Object testField0749 = new Integer(749);
    public Object testField0750 = new Integer(750);
    public Object testField0751 = new Integer(751);
    public Object testField0752 = new Integer(752);
    public Object testField0753 = new Integer(753);
    public Object testField0754 = new Integer(754);
    public Object testField0755 = new Integer(755);
    public Object testField0756 = new Integer(756);
    public Object testField0757 = new Integer(757);
    public Object testField0758 = new Integer(758);
    public Object testField0759 = new Integer(759);
    public Object testField0760 = new Integer(760);
    public Object testField0761 = new Integer(761);
    public Object testField0762 = new Integer(762);
    public Object testField0763 = new Integer(763);
    public Object testField0764 = new Integer(764);
    public Object testField0765 = new Integer(765);
    public Object testField0766 = new Integer(766);
    public Object testField0767 = new Integer(767);
    public Object testField0768 = new Integer(768);
    public Object testField0769 = new Integer(769);
    public Object testField0770 = new Integer(770);
    public Object testField0771 = new Integer(771);
    public Object testField0772 = new Integer(772);
    public Object testField0773 = new Integer(773);
    public Object testField0774 = new Integer(774);
    public Object testField0775 = new Integer(775);
    public Object testField0776 = new Integer(776);
    public Object testField0777 = new Integer(777);
    public Object testField0778 = new Integer(778);
    public Object testField0779 = new Integer(779);
    public Object testField0780 = new Integer(780);
    public Object testField0781 = new Integer(781);
    public Object testField0782 = new Integer(782);
    public Object testField0783 = new Integer(783);
    public Object testField0784 = new Integer(784);
    public Object testField0785 = new Integer(785);
    public Object testField0786 = new Integer(786);
    public Object testField0787 = new Integer(787);
    public Object testField0788 = new Integer(788);
    public Object testField0789 = new Integer(789);
    public Object testField0790 = new Integer(790);
    public Object testField0791 = new Integer(791);
    public Object testField0792 = new Integer(792);
    public Object testField0793 = new Integer(793);
    public Object testField0794 = new Integer(794);
    public Object testField0795 = new Integer(795);
    public Object testField0796 = new Integer(796);
    public Object testField0797 = new Integer(797);
    public Object testField0798 = new Integer(798);
    public Object testField0799 = new Integer(799);
    public Object testField0800 = new Integer(800);
    public Object testField0801 = new Integer(801);
    public Object testField0802 = new Integer(802);
    public Object testField0803 = new Integer(803);
    public Object testField0804 = new Integer(804);
    public Object testField0805 = new Integer(805);
    public Object testField0806 = new Integer(806);
    public Object testField0807 = new Integer(807);
    public Object testField0808 = new Integer(808);
    public Object testField0809 = new Integer(809);
    public Object testField0810 = new Integer(810);
    public Object testField0811 = new Integer(811);
    public Object testField0812 = new Integer(812);
    public Object testField0813 = new Integer(813);
    public Object testField0814 = new Integer(814);
    public Object testField0815 = new Integer(815);
    public Object testField0816 = new Integer(816);
    public Object testField0817 = new Integer(817);
    public Object testField0818 = new Integer(818);
    public Object testField0819 = new Integer(819);
    public Object testField0820 = new Integer(820);
    public Object testField0821 = new Integer(821);
    public Object testField0822 = new Integer(822);
    public Object testField0823 = new Integer(823);
    public Object testField0824 = new Integer(824);
    public Object testField0825 = new Integer(825);
    public Object testField0826 = new Integer(826);
    public Object testField0827 = new Integer(827);
    public Object testField0828 = new Integer(828);
    public Object testField0829 = new Integer(829);
    public Object testField0830 = new Integer(830);
    public Object testField0831 = new Integer(831);
    public Object testField0832 = new Integer(832);
    public Object testField0833 = new Integer(833);
    public Object testField0834 = new Integer(834);
    public Object testField0835 = new Integer(835);
    public Object testField0836 = new Integer(836);
    public Object testField0837 = new Integer(837);
    public Object testField0838 = new Integer(838);
    public Object testField0839 = new Integer(839);
    public Object testField0840 = new Integer(840);
    public Object testField0841 = new Integer(841);
    public Object testField0842 = new Integer(842);
    public Object testField0843 = new Integer(843);
    public Object testField0844 = new Integer(844);
    public Object testField0845 = new Integer(845);
    public Object testField0846 = new Integer(846);
    public Object testField0847 = new Integer(847);
    public Object testField0848 = new Integer(848);
    public Object testField0849 = new Integer(849);
    public Object testField0850 = new Integer(850);
    public Object testField0851 = new Integer(851);
    public Object testField0852 = new Integer(852);
    public Object testField0853 = new Integer(853);
    public Object testField0854 = new Integer(854);
    public Object testField0855 = new Integer(855);
    public Object testField0856 = new Integer(856);
    public Object testField0857 = new Integer(857);
    public Object testField0858 = new Integer(858);
    public Object testField0859 = new Integer(859);
    public Object testField0860 = new Integer(860);
    public Object testField0861 = new Integer(861);
    public Object testField0862 = new Integer(862);
    public Object testField0863 = new Integer(863);
    public Object testField0864 = new Integer(864);
    public Object testField0865 = new Integer(865);
    public Object testField0866 = new Integer(866);
    public Object testField0867 = new Integer(867);
    public Object testField0868 = new Integer(868);
    public Object testField0869 = new Integer(869);
    public Object testField0870 = new Integer(870);
    public Object testField0871 = new Integer(871);
    public Object testField0872 = new Integer(872);
    public Object testField0873 = new Integer(873);
    public Object testField0874 = new Integer(874);
    public Object testField0875 = new Integer(875);
    public Object testField0876 = new Integer(876);
    public Object testField0877 = new Integer(877);
    public Object testField0878 = new Integer(878);
    public Object testField0879 = new Integer(879);
    public Object testField0880 = new Integer(880);
    public Object testField0881 = new Integer(881);
    public Object testField0882 = new Integer(882);
    public Object testField0883 = new Integer(883);
    public Object testField0884 = new Integer(884);
    public Object testField0885 = new Integer(885);
    public Object testField0886 = new Integer(886);
    public Object testField0887 = new Integer(887);
    public Object testField0888 = new Integer(888);
    public Object testField0889 = new Integer(889);
    public Object testField0890 = new Integer(890);
    public Object testField0891 = new Integer(891);
    public Object testField0892 = new Integer(892);
    public Object testField0893 = new Integer(893);
    public Object testField0894 = new Integer(894);
    public Object testField0895 = new Integer(895);
    public Object testField0896 = new Integer(896);
    public Object testField0897 = new Integer(897);
    public Object testField0898 = new Integer(898);
    public Object testField0899 = new Integer(899);
    public Object testField0900 = new Integer(900);
    public Object testField0901 = new Integer(901);
    public Object testField0902 = new Integer(902);
    public Object testField0903 = new Integer(903);
    public Object testField0904 = new Integer(904);
    public Object testField0905 = new Integer(905);
    public Object testField0906 = new Integer(906);
    public Object testField0907 = new Integer(907);
    public Object testField0908 = new Integer(908);
    public Object testField0909 = new Integer(909);
    public Object testField0910 = new Integer(910);
    public Object testField0911 = new Integer(911);
    public Object testField0912 = new Integer(912);
    public Object testField0913 = new Integer(913);
    public Object testField0914 = new Integer(914);
    public Object testField0915 = new Integer(915);
    public Object testField0916 = new Integer(916);
    public Object testField0917 = new Integer(917);
    public Object testField0918 = new Integer(918);
    public Object testField0919 = new Integer(919);
    public Object testField0920 = new Integer(920);
    public Object testField0921 = new Integer(921);
    public Object testField0922 = new Integer(922);
    public Object testField0923 = new Integer(923);
    public Object testField0924 = new Integer(924);
    public Object testField0925 = new Integer(925);
    public Object testField0926 = new Integer(926);
    public Object testField0927 = new Integer(927);
    public Object testField0928 = new Integer(928);
    public Object testField0929 = new Integer(929);
    public Object testField0930 = new Integer(930);
    public Object testField0931 = new Integer(931);
    public Object testField0932 = new Integer(932);
    public Object testField0933 = new Integer(933);
    public Object testField0934 = new Integer(934);
    public Object testField0935 = new Integer(935);
    public Object testField0936 = new Integer(936);
    public Object testField0937 = new Integer(937);
    public Object testField0938 = new Integer(938);
    public Object testField0939 = new Integer(939);
    public Object testField0940 = new Integer(940);
    public Object testField0941 = new Integer(941);
    public Object testField0942 = new Integer(942);
    public Object testField0943 = new Integer(943);
    public Object testField0944 = new Integer(944);
    public Object testField0945 = new Integer(945);
    public Object testField0946 = new Integer(946);
    public Object testField0947 = new Integer(947);
    public Object testField0948 = new Integer(948);
    public Object testField0949 = new Integer(949);
    public Object testField0950 = new Integer(950);
    public Object testField0951 = new Integer(951);
    public Object testField0952 = new Integer(952);
    public Object testField0953 = new Integer(953);
    public Object testField0954 = new Integer(954);
    public Object testField0955 = new Integer(955);
    public Object testField0956 = new Integer(956);
    public Object testField0957 = new Integer(957);
    public Object testField0958 = new Integer(958);
    public Object testField0959 = new Integer(959);
    public Object testField0960 = new Integer(960);
    public Object testField0961 = new Integer(961);
    public Object testField0962 = new Integer(962);
    public Object testField0963 = new Integer(963);
    public Object testField0964 = new Integer(964);
    public Object testField0965 = new Integer(965);
    public Object testField0966 = new Integer(966);
    public Object testField0967 = new Integer(967);
    public Object testField0968 = new Integer(968);
    public Object testField0969 = new Integer(969);
    public Object testField0970 = new Integer(970);
    public Object testField0971 = new Integer(971);
    public Object testField0972 = new Integer(972);
    public Object testField0973 = new Integer(973);
    public Object testField0974 = new Integer(974);
    public Object testField0975 = new Integer(975);
    public Object testField0976 = new Integer(976);
    public Object testField0977 = new Integer(977);
    public Object testField0978 = new Integer(978);
    public Object testField0979 = new Integer(979);
    public Object testField0980 = new Integer(980);
    public Object testField0981 = new Integer(981);
    public Object testField0982 = new Integer(982);
    public Object testField0983 = new Integer(983);
    public Object testField0984 = new Integer(984);
    public Object testField0985 = new Integer(985);
    public Object testField0986 = new Integer(986);
    public Object testField0987 = new Integer(987);
    public Object testField0988 = new Integer(988);
    public Object testField0989 = new Integer(989);
    public Object testField0990 = new Integer(990);
    public Object testField0991 = new Integer(991);
    public Object testField0992 = new Integer(992);
    public Object testField0993 = new Integer(993);
    public Object testField0994 = new Integer(994);
    public Object testField0995 = new Integer(995);
    public Object testField0996 = new Integer(996);
    public Object testField0997 = new Integer(997);
    public Object testField0998 = new Integer(998);
    public Object testField0999 = new Integer(999);
}
