/*
 * Copyright (C) 2012 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
  public static void main(String args[]) throws Exception {
    p1.BaseClass c = new p2.DerivedClass();
    c.run();
    p2.DerivedClass d = new p2.DerivedClass();
    d.bar();
    p2.DerivedClass d2 = new p2.DerivedClass2();
    d2.bar();
    p2.DerivedClass2 d3 = new p2.DerivedClass2();
    d3.bar();
  }
}
