/*
 * Copyright (C) 2015 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
  // A very simple check that disassembly information has been added to the
  // graph. We check that sections have been added for the frame entry and a
  // slow path.
  /// CHECK-START: int Main.DisassembledFunction(int) disassembly (after)
  /// CHECK:       FrameEntry
  /// CHECK:       DivZeroCheckSlowPath{{.*}}
  public int DisassembledFunction(int arg) {
    return 7 / arg;
  }

  public static void main(String[] args) {}
}
