/*
 * Copyright (C) 2016 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

interface IFace {
  public default void $noinline$aMethod() { throw new RuntimeException("Should not be called"); }
}

class ClassImplA implements IFace {
  /// CHECK-START: void ClassImplA.testSuperInvoke() builder (after)
  /// CHECK:                       InvokeStaticOrDirect
  public void testSuperInvoke() {
    IFace.super.$noinline$aMethod();
  }
}

class ClassImplB extends ClassImplA {
  /// CHECK-START: void ClassImplB.testSuperInvoke2() builder (after)
  /// CHECK:                       InvokeStaticOrDirect
  public void testSuperInvoke2() {
    super.$noinline$aMethod();
  }
}

public class Main {
  public static void main(String[] args) { }
}
