/*
 * Copyright (C) 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License
 */

package com.android.class2greylist;

public class JavadocLinkSyntaxError extends Exception {
    public final String expected;
    public final int position;
    public final String context;

    public JavadocLinkSyntaxError(String expected, StringCursor sc) {
        super(expected + " at position " + sc.position() + " in " + sc.getOriginalString());
        this.expected = expected;
        this.position = sc.position();
        this.context = sc.getOriginalString();
    }
}

