/*
 * Copyright (C) 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "base/locks.h"

/*
 * Stub definitions for targets without nterp implementations.
 */

namespace art {

class ArtMethod;

namespace interpreter {

bool IsNterpSupported() {
  return false;
}

bool CanRuntimeUseNterp() {
  return false;
}

bool CanMethodUseNterp(ArtMethod* method ATTRIBUTE_UNUSED) {
  return false;
}

const void* GetNterpEntryPoint() {
  return nullptr;
}

void CheckNterpAsmConstants() {
}

extern "C" void ExecuteNterpImpl() REQUIRES_SHARED(Locks::mutator_lock_) {
  UNIMPLEMENTED(FATAL);
}

extern "C" void* artNterpAsmInstructionStart[] = { nullptr };
extern "C" void* artNterpAsmInstructionEnd[] = { nullptr };

}  // namespace interpreter
}  // namespace art
