/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.ahat.heapdump;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * An iterator that skips over nulls.
 */
class SkipNullsIterator<T> implements Iterator<T>, Iterable<T> {
  Iterator<T> mIter;
  private T mNext;

  public SkipNullsIterator(Iterable<T> iterable) {
    mIter = iterable.iterator();
    mNext = null;
  }

  @Override
  public boolean hasNext() {
    while (mNext == null && mIter.hasNext()) {
      mNext = mIter.next();
    }
    return mNext != null;
  }

  @Override
  public T next() {
    if (!hasNext()) {
      throw new NoSuchElementException();
    }
    T next = mNext;
    mNext = null;
    return next;
  }

  @Override
  public Iterator<T> iterator() {
    return this;
  }
}
