/*
 * Copyright (C) 2014 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Test extends Super {
    public void test(Test t) {
        t.print();
    }

    private void print() {
        System.out.println("Test");
    }

    /// CHECK-START: java.lang.Integer Test.toInteger() builder (after)
    /// CHECK:         LoadClass needs_access_check:false klass:java.lang.Integer

    public Integer toInteger() {
        return new Integer(42);
    }

    public String toString() {
        return new String("Test");
    }
}
