/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package art.constmethodhandle;

import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodType;

public class TestInvoke extends BaseTestInvoke {
  // THIS IS GENERATED BY ASM
  // @Override
  // public void runTest(Runnable preCall) {
  //   preCall.run();
  //   // MAGIC! Replaced with a static method handle pointing at
  //   // art.constmethodhandle.Responses.sayHi // (or art.constmethodhandle.Responses.sayBye in the
  //   // redefined version).
  //   MethodHandle handle = CONSTANT_MH<Responses.sayHi>;
  //   handle.invokeExact();
  // }
}
