/*
 * Copyright (C) 2011 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

class Statics {
    static final boolean s0 = true;
    static final byte s1 = 5;
    static final char s2 = 'a';
    static final short s3 = (short) 65000;
    static final int s4 = 2000000000;
    static final long s5 = 0x1234567890abcdefL;
    static final float s6 = 0.5f;
    static final double s7 = 16777217;
    static final String s8 = "android";

    static boolean getS0() {
        return s0;
    }
    static byte getS1() {
        return s1;
    }
    static char getS2() {
        return s2;
    }
    static short getS3() {
        return s3;
    }
    static int getS4() {
        return s4;
    }
    static long getS5() {
        return s5;
    }
    static float getS6() {
        return s6;
    }
    static double getS7() {
        return s7;
    }
    static String getS8() {
        return s8;
    }
}
