/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

class ManyFieldsBase1 extends ManyFieldsBase0 {
    public Object testField1000 = new Integer(1000);
    public Object testField1001 = new Integer(1001);
    public Object testField1002 = new Integer(1002);
    public Object testField1003 = new Integer(1003);
    public Object testField1004 = new Integer(1004);
    public Object testField1005 = new Integer(1005);
    public Object testField1006 = new Integer(1006);
    public Object testField1007 = new Integer(1007);
    public Object testField1008 = new Integer(1008);
    public Object testField1009 = new Integer(1009);
    public Object testField1010 = new Integer(1010);
    public Object testField1011 = new Integer(1011);
    public Object testField1012 = new Integer(1012);
    public Object testField1013 = new Integer(1013);
    public Object testField1014 = new Integer(1014);
    public Object testField1015 = new Integer(1015);
    public Object testField1016 = new Integer(1016);
    public Object testField1017 = new Integer(1017);
    public Object testField1018 = new Integer(1018);
    public Object testField1019 = new Integer(1019);
    public Object testField1020 = new Integer(1020);
    public Object testField1021 = new Integer(1021);
    public Object testField1022 = new Integer(1022);
    public Object testField1023 = new Integer(1023);
    public Object testField1024 = new Integer(1024);
    public Object testField1025 = new Integer(1025);
    public Object testField1026 = new Integer(1026);
    public Object testField1027 = new Integer(1027);
    public Object testField1028 = new Integer(1028);
    public Object testField1029 = new Integer(1029);
    public Object testField1030 = new Integer(1030);
    public Object testField1031 = new Integer(1031);
    public Object testField1032 = new Integer(1032);
    public Object testField1033 = new Integer(1033);
    public Object testField1034 = new Integer(1034);
    public Object testField1035 = new Integer(1035);
    public Object testField1036 = new Integer(1036);
    public Object testField1037 = new Integer(1037);
    public Object testField1038 = new Integer(1038);
    public Object testField1039 = new Integer(1039);
    public Object testField1040 = new Integer(1040);
    public Object testField1041 = new Integer(1041);
    public Object testField1042 = new Integer(1042);
    public Object testField1043 = new Integer(1043);
    public Object testField1044 = new Integer(1044);
    public Object testField1045 = new Integer(1045);
    public Object testField1046 = new Integer(1046);
    public Object testField1047 = new Integer(1047);
    public Object testField1048 = new Integer(1048);
    public Object testField1049 = new Integer(1049);
    public Object testField1050 = new Integer(1050);
    public Object testField1051 = new Integer(1051);
    public Object testField1052 = new Integer(1052);
    public Object testField1053 = new Integer(1053);
    public Object testField1054 = new Integer(1054);
    public Object testField1055 = new Integer(1055);
    public Object testField1056 = new Integer(1056);
    public Object testField1057 = new Integer(1057);
    public Object testField1058 = new Integer(1058);
    public Object testField1059 = new Integer(1059);
    public Object testField1060 = new Integer(1060);
    public Object testField1061 = new Integer(1061);
    public Object testField1062 = new Integer(1062);
    public Object testField1063 = new Integer(1063);
    public Object testField1064 = new Integer(1064);
    public Object testField1065 = new Integer(1065);
    public Object testField1066 = new Integer(1066);
    public Object testField1067 = new Integer(1067);
    public Object testField1068 = new Integer(1068);
    public Object testField1069 = new Integer(1069);
    public Object testField1070 = new Integer(1070);
    public Object testField1071 = new Integer(1071);
    public Object testField1072 = new Integer(1072);
    public Object testField1073 = new Integer(1073);
    public Object testField1074 = new Integer(1074);
    public Object testField1075 = new Integer(1075);
    public Object testField1076 = new Integer(1076);
    public Object testField1077 = new Integer(1077);
    public Object testField1078 = new Integer(1078);
    public Object testField1079 = new Integer(1079);
    public Object testField1080 = new Integer(1080);
    public Object testField1081 = new Integer(1081);
    public Object testField1082 = new Integer(1082);
    public Object testField1083 = new Integer(1083);
    public Object testField1084 = new Integer(1084);
    public Object testField1085 = new Integer(1085);
    public Object testField1086 = new Integer(1086);
    public Object testField1087 = new Integer(1087);
    public Object testField1088 = new Integer(1088);
    public Object testField1089 = new Integer(1089);
    public Object testField1090 = new Integer(1090);
    public Object testField1091 = new Integer(1091);
    public Object testField1092 = new Integer(1092);
    public Object testField1093 = new Integer(1093);
    public Object testField1094 = new Integer(1094);
    public Object testField1095 = new Integer(1095);
    public Object testField1096 = new Integer(1096);
    public Object testField1097 = new Integer(1097);
    public Object testField1098 = new Integer(1098);
    public Object testField1099 = new Integer(1099);
    public Object testField1100 = new Integer(1100);
    public Object testField1101 = new Integer(1101);
    public Object testField1102 = new Integer(1102);
    public Object testField1103 = new Integer(1103);
    public Object testField1104 = new Integer(1104);
    public Object testField1105 = new Integer(1105);
    public Object testField1106 = new Integer(1106);
    public Object testField1107 = new Integer(1107);
    public Object testField1108 = new Integer(1108);
    public Object testField1109 = new Integer(1109);
    public Object testField1110 = new Integer(1110);
    public Object testField1111 = new Integer(1111);
    public Object testField1112 = new Integer(1112);
    public Object testField1113 = new Integer(1113);
    public Object testField1114 = new Integer(1114);
    public Object testField1115 = new Integer(1115);
    public Object testField1116 = new Integer(1116);
    public Object testField1117 = new Integer(1117);
    public Object testField1118 = new Integer(1118);
    public Object testField1119 = new Integer(1119);
    public Object testField1120 = new Integer(1120);
    public Object testField1121 = new Integer(1121);
    public Object testField1122 = new Integer(1122);
    public Object testField1123 = new Integer(1123);
    public Object testField1124 = new Integer(1124);
    public Object testField1125 = new Integer(1125);
    public Object testField1126 = new Integer(1126);
    public Object testField1127 = new Integer(1127);
    public Object testField1128 = new Integer(1128);
    public Object testField1129 = new Integer(1129);
    public Object testField1130 = new Integer(1130);
    public Object testField1131 = new Integer(1131);
    public Object testField1132 = new Integer(1132);
    public Object testField1133 = new Integer(1133);
    public Object testField1134 = new Integer(1134);
    public Object testField1135 = new Integer(1135);
    public Object testField1136 = new Integer(1136);
    public Object testField1137 = new Integer(1137);
    public Object testField1138 = new Integer(1138);
    public Object testField1139 = new Integer(1139);
    public Object testField1140 = new Integer(1140);
    public Object testField1141 = new Integer(1141);
    public Object testField1142 = new Integer(1142);
    public Object testField1143 = new Integer(1143);
    public Object testField1144 = new Integer(1144);
    public Object testField1145 = new Integer(1145);
    public Object testField1146 = new Integer(1146);
    public Object testField1147 = new Integer(1147);
    public Object testField1148 = new Integer(1148);
    public Object testField1149 = new Integer(1149);
    public Object testField1150 = new Integer(1150);
    public Object testField1151 = new Integer(1151);
    public Object testField1152 = new Integer(1152);
    public Object testField1153 = new Integer(1153);
    public Object testField1154 = new Integer(1154);
    public Object testField1155 = new Integer(1155);
    public Object testField1156 = new Integer(1156);
    public Object testField1157 = new Integer(1157);
    public Object testField1158 = new Integer(1158);
    public Object testField1159 = new Integer(1159);
    public Object testField1160 = new Integer(1160);
    public Object testField1161 = new Integer(1161);
    public Object testField1162 = new Integer(1162);
    public Object testField1163 = new Integer(1163);
    public Object testField1164 = new Integer(1164);
    public Object testField1165 = new Integer(1165);
    public Object testField1166 = new Integer(1166);
    public Object testField1167 = new Integer(1167);
    public Object testField1168 = new Integer(1168);
    public Object testField1169 = new Integer(1169);
    public Object testField1170 = new Integer(1170);
    public Object testField1171 = new Integer(1171);
    public Object testField1172 = new Integer(1172);
    public Object testField1173 = new Integer(1173);
    public Object testField1174 = new Integer(1174);
    public Object testField1175 = new Integer(1175);
    public Object testField1176 = new Integer(1176);
    public Object testField1177 = new Integer(1177);
    public Object testField1178 = new Integer(1178);
    public Object testField1179 = new Integer(1179);
    public Object testField1180 = new Integer(1180);
    public Object testField1181 = new Integer(1181);
    public Object testField1182 = new Integer(1182);
    public Object testField1183 = new Integer(1183);
    public Object testField1184 = new Integer(1184);
    public Object testField1185 = new Integer(1185);
    public Object testField1186 = new Integer(1186);
    public Object testField1187 = new Integer(1187);
    public Object testField1188 = new Integer(1188);
    public Object testField1189 = new Integer(1189);
    public Object testField1190 = new Integer(1190);
    public Object testField1191 = new Integer(1191);
    public Object testField1192 = new Integer(1192);
    public Object testField1193 = new Integer(1193);
    public Object testField1194 = new Integer(1194);
    public Object testField1195 = new Integer(1195);
    public Object testField1196 = new Integer(1196);
    public Object testField1197 = new Integer(1197);
    public Object testField1198 = new Integer(1198);
    public Object testField1199 = new Integer(1199);
    public Object testField1200 = new Integer(1200);
    public Object testField1201 = new Integer(1201);
    public Object testField1202 = new Integer(1202);
    public Object testField1203 = new Integer(1203);
    public Object testField1204 = new Integer(1204);
    public Object testField1205 = new Integer(1205);
    public Object testField1206 = new Integer(1206);
    public Object testField1207 = new Integer(1207);
    public Object testField1208 = new Integer(1208);
    public Object testField1209 = new Integer(1209);
    public Object testField1210 = new Integer(1210);
    public Object testField1211 = new Integer(1211);
    public Object testField1212 = new Integer(1212);
    public Object testField1213 = new Integer(1213);
    public Object testField1214 = new Integer(1214);
    public Object testField1215 = new Integer(1215);
    public Object testField1216 = new Integer(1216);
    public Object testField1217 = new Integer(1217);
    public Object testField1218 = new Integer(1218);
    public Object testField1219 = new Integer(1219);
    public Object testField1220 = new Integer(1220);
    public Object testField1221 = new Integer(1221);
    public Object testField1222 = new Integer(1222);
    public Object testField1223 = new Integer(1223);
    public Object testField1224 = new Integer(1224);
    public Object testField1225 = new Integer(1225);
    public Object testField1226 = new Integer(1226);
    public Object testField1227 = new Integer(1227);
    public Object testField1228 = new Integer(1228);
    public Object testField1229 = new Integer(1229);
    public Object testField1230 = new Integer(1230);
    public Object testField1231 = new Integer(1231);
    public Object testField1232 = new Integer(1232);
    public Object testField1233 = new Integer(1233);
    public Object testField1234 = new Integer(1234);
    public Object testField1235 = new Integer(1235);
    public Object testField1236 = new Integer(1236);
    public Object testField1237 = new Integer(1237);
    public Object testField1238 = new Integer(1238);
    public Object testField1239 = new Integer(1239);
    public Object testField1240 = new Integer(1240);
    public Object testField1241 = new Integer(1241);
    public Object testField1242 = new Integer(1242);
    public Object testField1243 = new Integer(1243);
    public Object testField1244 = new Integer(1244);
    public Object testField1245 = new Integer(1245);
    public Object testField1246 = new Integer(1246);
    public Object testField1247 = new Integer(1247);
    public Object testField1248 = new Integer(1248);
    public Object testField1249 = new Integer(1249);
    public Object testField1250 = new Integer(1250);
    public Object testField1251 = new Integer(1251);
    public Object testField1252 = new Integer(1252);
    public Object testField1253 = new Integer(1253);
    public Object testField1254 = new Integer(1254);
    public Object testField1255 = new Integer(1255);
    public Object testField1256 = new Integer(1256);
    public Object testField1257 = new Integer(1257);
    public Object testField1258 = new Integer(1258);
    public Object testField1259 = new Integer(1259);
    public Object testField1260 = new Integer(1260);
    public Object testField1261 = new Integer(1261);
    public Object testField1262 = new Integer(1262);
    public Object testField1263 = new Integer(1263);
    public Object testField1264 = new Integer(1264);
    public Object testField1265 = new Integer(1265);
    public Object testField1266 = new Integer(1266);
    public Object testField1267 = new Integer(1267);
    public Object testField1268 = new Integer(1268);
    public Object testField1269 = new Integer(1269);
    public Object testField1270 = new Integer(1270);
    public Object testField1271 = new Integer(1271);
    public Object testField1272 = new Integer(1272);
    public Object testField1273 = new Integer(1273);
    public Object testField1274 = new Integer(1274);
    public Object testField1275 = new Integer(1275);
    public Object testField1276 = new Integer(1276);
    public Object testField1277 = new Integer(1277);
    public Object testField1278 = new Integer(1278);
    public Object testField1279 = new Integer(1279);
    public Object testField1280 = new Integer(1280);
    public Object testField1281 = new Integer(1281);
    public Object testField1282 = new Integer(1282);
    public Object testField1283 = new Integer(1283);
    public Object testField1284 = new Integer(1284);
    public Object testField1285 = new Integer(1285);
    public Object testField1286 = new Integer(1286);
    public Object testField1287 = new Integer(1287);
    public Object testField1288 = new Integer(1288);
    public Object testField1289 = new Integer(1289);
    public Object testField1290 = new Integer(1290);
    public Object testField1291 = new Integer(1291);
    public Object testField1292 = new Integer(1292);
    public Object testField1293 = new Integer(1293);
    public Object testField1294 = new Integer(1294);
    public Object testField1295 = new Integer(1295);
    public Object testField1296 = new Integer(1296);
    public Object testField1297 = new Integer(1297);
    public Object testField1298 = new Integer(1298);
    public Object testField1299 = new Integer(1299);
    public Object testField1300 = new Integer(1300);
    public Object testField1301 = new Integer(1301);
    public Object testField1302 = new Integer(1302);
    public Object testField1303 = new Integer(1303);
    public Object testField1304 = new Integer(1304);
    public Object testField1305 = new Integer(1305);
    public Object testField1306 = new Integer(1306);
    public Object testField1307 = new Integer(1307);
    public Object testField1308 = new Integer(1308);
    public Object testField1309 = new Integer(1309);
    public Object testField1310 = new Integer(1310);
    public Object testField1311 = new Integer(1311);
    public Object testField1312 = new Integer(1312);
    public Object testField1313 = new Integer(1313);
    public Object testField1314 = new Integer(1314);
    public Object testField1315 = new Integer(1315);
    public Object testField1316 = new Integer(1316);
    public Object testField1317 = new Integer(1317);
    public Object testField1318 = new Integer(1318);
    public Object testField1319 = new Integer(1319);
    public Object testField1320 = new Integer(1320);
    public Object testField1321 = new Integer(1321);
    public Object testField1322 = new Integer(1322);
    public Object testField1323 = new Integer(1323);
    public Object testField1324 = new Integer(1324);
    public Object testField1325 = new Integer(1325);
    public Object testField1326 = new Integer(1326);
    public Object testField1327 = new Integer(1327);
    public Object testField1328 = new Integer(1328);
    public Object testField1329 = new Integer(1329);
    public Object testField1330 = new Integer(1330);
    public Object testField1331 = new Integer(1331);
    public Object testField1332 = new Integer(1332);
    public Object testField1333 = new Integer(1333);
    public Object testField1334 = new Integer(1334);
    public Object testField1335 = new Integer(1335);
    public Object testField1336 = new Integer(1336);
    public Object testField1337 = new Integer(1337);
    public Object testField1338 = new Integer(1338);
    public Object testField1339 = new Integer(1339);
    public Object testField1340 = new Integer(1340);
    public Object testField1341 = new Integer(1341);
    public Object testField1342 = new Integer(1342);
    public Object testField1343 = new Integer(1343);
    public Object testField1344 = new Integer(1344);
    public Object testField1345 = new Integer(1345);
    public Object testField1346 = new Integer(1346);
    public Object testField1347 = new Integer(1347);
    public Object testField1348 = new Integer(1348);
    public Object testField1349 = new Integer(1349);
    public Object testField1350 = new Integer(1350);
    public Object testField1351 = new Integer(1351);
    public Object testField1352 = new Integer(1352);
    public Object testField1353 = new Integer(1353);
    public Object testField1354 = new Integer(1354);
    public Object testField1355 = new Integer(1355);
    public Object testField1356 = new Integer(1356);
    public Object testField1357 = new Integer(1357);
    public Object testField1358 = new Integer(1358);
    public Object testField1359 = new Integer(1359);
    public Object testField1360 = new Integer(1360);
    public Object testField1361 = new Integer(1361);
    public Object testField1362 = new Integer(1362);
    public Object testField1363 = new Integer(1363);
    public Object testField1364 = new Integer(1364);
    public Object testField1365 = new Integer(1365);
    public Object testField1366 = new Integer(1366);
    public Object testField1367 = new Integer(1367);
    public Object testField1368 = new Integer(1368);
    public Object testField1369 = new Integer(1369);
    public Object testField1370 = new Integer(1370);
    public Object testField1371 = new Integer(1371);
    public Object testField1372 = new Integer(1372);
    public Object testField1373 = new Integer(1373);
    public Object testField1374 = new Integer(1374);
    public Object testField1375 = new Integer(1375);
    public Object testField1376 = new Integer(1376);
    public Object testField1377 = new Integer(1377);
    public Object testField1378 = new Integer(1378);
    public Object testField1379 = new Integer(1379);
    public Object testField1380 = new Integer(1380);
    public Object testField1381 = new Integer(1381);
    public Object testField1382 = new Integer(1382);
    public Object testField1383 = new Integer(1383);
    public Object testField1384 = new Integer(1384);
    public Object testField1385 = new Integer(1385);
    public Object testField1386 = new Integer(1386);
    public Object testField1387 = new Integer(1387);
    public Object testField1388 = new Integer(1388);
    public Object testField1389 = new Integer(1389);
    public Object testField1390 = new Integer(1390);
    public Object testField1391 = new Integer(1391);
    public Object testField1392 = new Integer(1392);
    public Object testField1393 = new Integer(1393);
    public Object testField1394 = new Integer(1394);
    public Object testField1395 = new Integer(1395);
    public Object testField1396 = new Integer(1396);
    public Object testField1397 = new Integer(1397);
    public Object testField1398 = new Integer(1398);
    public Object testField1399 = new Integer(1399);
    public Object testField1400 = new Integer(1400);
    public Object testField1401 = new Integer(1401);
    public Object testField1402 = new Integer(1402);
    public Object testField1403 = new Integer(1403);
    public Object testField1404 = new Integer(1404);
    public Object testField1405 = new Integer(1405);
    public Object testField1406 = new Integer(1406);
    public Object testField1407 = new Integer(1407);
    public Object testField1408 = new Integer(1408);
    public Object testField1409 = new Integer(1409);
    public Object testField1410 = new Integer(1410);
    public Object testField1411 = new Integer(1411);
    public Object testField1412 = new Integer(1412);
    public Object testField1413 = new Integer(1413);
    public Object testField1414 = new Integer(1414);
    public Object testField1415 = new Integer(1415);
    public Object testField1416 = new Integer(1416);
    public Object testField1417 = new Integer(1417);
    public Object testField1418 = new Integer(1418);
    public Object testField1419 = new Integer(1419);
    public Object testField1420 = new Integer(1420);
    public Object testField1421 = new Integer(1421);
    public Object testField1422 = new Integer(1422);
    public Object testField1423 = new Integer(1423);
    public Object testField1424 = new Integer(1424);
    public Object testField1425 = new Integer(1425);
    public Object testField1426 = new Integer(1426);
    public Object testField1427 = new Integer(1427);
    public Object testField1428 = new Integer(1428);
    public Object testField1429 = new Integer(1429);
    public Object testField1430 = new Integer(1430);
    public Object testField1431 = new Integer(1431);
    public Object testField1432 = new Integer(1432);
    public Object testField1433 = new Integer(1433);
    public Object testField1434 = new Integer(1434);
    public Object testField1435 = new Integer(1435);
    public Object testField1436 = new Integer(1436);
    public Object testField1437 = new Integer(1437);
    public Object testField1438 = new Integer(1438);
    public Object testField1439 = new Integer(1439);
    public Object testField1440 = new Integer(1440);
    public Object testField1441 = new Integer(1441);
    public Object testField1442 = new Integer(1442);
    public Object testField1443 = new Integer(1443);
    public Object testField1444 = new Integer(1444);
    public Object testField1445 = new Integer(1445);
    public Object testField1446 = new Integer(1446);
    public Object testField1447 = new Integer(1447);
    public Object testField1448 = new Integer(1448);
    public Object testField1449 = new Integer(1449);
    public Object testField1450 = new Integer(1450);
    public Object testField1451 = new Integer(1451);
    public Object testField1452 = new Integer(1452);
    public Object testField1453 = new Integer(1453);
    public Object testField1454 = new Integer(1454);
    public Object testField1455 = new Integer(1455);
    public Object testField1456 = new Integer(1456);
    public Object testField1457 = new Integer(1457);
    public Object testField1458 = new Integer(1458);
    public Object testField1459 = new Integer(1459);
    public Object testField1460 = new Integer(1460);
    public Object testField1461 = new Integer(1461);
    public Object testField1462 = new Integer(1462);
    public Object testField1463 = new Integer(1463);
    public Object testField1464 = new Integer(1464);
    public Object testField1465 = new Integer(1465);
    public Object testField1466 = new Integer(1466);
    public Object testField1467 = new Integer(1467);
    public Object testField1468 = new Integer(1468);
    public Object testField1469 = new Integer(1469);
    public Object testField1470 = new Integer(1470);
    public Object testField1471 = new Integer(1471);
    public Object testField1472 = new Integer(1472);
    public Object testField1473 = new Integer(1473);
    public Object testField1474 = new Integer(1474);
    public Object testField1475 = new Integer(1475);
    public Object testField1476 = new Integer(1476);
    public Object testField1477 = new Integer(1477);
    public Object testField1478 = new Integer(1478);
    public Object testField1479 = new Integer(1479);
    public Object testField1480 = new Integer(1480);
    public Object testField1481 = new Integer(1481);
    public Object testField1482 = new Integer(1482);
    public Object testField1483 = new Integer(1483);
    public Object testField1484 = new Integer(1484);
    public Object testField1485 = new Integer(1485);
    public Object testField1486 = new Integer(1486);
    public Object testField1487 = new Integer(1487);
    public Object testField1488 = new Integer(1488);
    public Object testField1489 = new Integer(1489);
    public Object testField1490 = new Integer(1490);
    public Object testField1491 = new Integer(1491);
    public Object testField1492 = new Integer(1492);
    public Object testField1493 = new Integer(1493);
    public Object testField1494 = new Integer(1494);
    public Object testField1495 = new Integer(1495);
    public Object testField1496 = new Integer(1496);
    public Object testField1497 = new Integer(1497);
    public Object testField1498 = new Integer(1498);
    public Object testField1499 = new Integer(1499);
    public Object testField1500 = new Integer(1500);
    public Object testField1501 = new Integer(1501);
    public Object testField1502 = new Integer(1502);
    public Object testField1503 = new Integer(1503);
    public Object testField1504 = new Integer(1504);
    public Object testField1505 = new Integer(1505);
    public Object testField1506 = new Integer(1506);
    public Object testField1507 = new Integer(1507);
    public Object testField1508 = new Integer(1508);
    public Object testField1509 = new Integer(1509);
    public Object testField1510 = new Integer(1510);
    public Object testField1511 = new Integer(1511);
    public Object testField1512 = new Integer(1512);
    public Object testField1513 = new Integer(1513);
    public Object testField1514 = new Integer(1514);
    public Object testField1515 = new Integer(1515);
    public Object testField1516 = new Integer(1516);
    public Object testField1517 = new Integer(1517);
    public Object testField1518 = new Integer(1518);
    public Object testField1519 = new Integer(1519);
    public Object testField1520 = new Integer(1520);
    public Object testField1521 = new Integer(1521);
    public Object testField1522 = new Integer(1522);
    public Object testField1523 = new Integer(1523);
    public Object testField1524 = new Integer(1524);
    public Object testField1525 = new Integer(1525);
    public Object testField1526 = new Integer(1526);
    public Object testField1527 = new Integer(1527);
    public Object testField1528 = new Integer(1528);
    public Object testField1529 = new Integer(1529);
    public Object testField1530 = new Integer(1530);
    public Object testField1531 = new Integer(1531);
    public Object testField1532 = new Integer(1532);
    public Object testField1533 = new Integer(1533);
    public Object testField1534 = new Integer(1534);
    public Object testField1535 = new Integer(1535);
    public Object testField1536 = new Integer(1536);
    public Object testField1537 = new Integer(1537);
    public Object testField1538 = new Integer(1538);
    public Object testField1539 = new Integer(1539);
    public Object testField1540 = new Integer(1540);
    public Object testField1541 = new Integer(1541);
    public Object testField1542 = new Integer(1542);
    public Object testField1543 = new Integer(1543);
    public Object testField1544 = new Integer(1544);
    public Object testField1545 = new Integer(1545);
    public Object testField1546 = new Integer(1546);
    public Object testField1547 = new Integer(1547);
    public Object testField1548 = new Integer(1548);
    public Object testField1549 = new Integer(1549);
    public Object testField1550 = new Integer(1550);
    public Object testField1551 = new Integer(1551);
    public Object testField1552 = new Integer(1552);
    public Object testField1553 = new Integer(1553);
    public Object testField1554 = new Integer(1554);
    public Object testField1555 = new Integer(1555);
    public Object testField1556 = new Integer(1556);
    public Object testField1557 = new Integer(1557);
    public Object testField1558 = new Integer(1558);
    public Object testField1559 = new Integer(1559);
    public Object testField1560 = new Integer(1560);
    public Object testField1561 = new Integer(1561);
    public Object testField1562 = new Integer(1562);
    public Object testField1563 = new Integer(1563);
    public Object testField1564 = new Integer(1564);
    public Object testField1565 = new Integer(1565);
    public Object testField1566 = new Integer(1566);
    public Object testField1567 = new Integer(1567);
    public Object testField1568 = new Integer(1568);
    public Object testField1569 = new Integer(1569);
    public Object testField1570 = new Integer(1570);
    public Object testField1571 = new Integer(1571);
    public Object testField1572 = new Integer(1572);
    public Object testField1573 = new Integer(1573);
    public Object testField1574 = new Integer(1574);
    public Object testField1575 = new Integer(1575);
    public Object testField1576 = new Integer(1576);
    public Object testField1577 = new Integer(1577);
    public Object testField1578 = new Integer(1578);
    public Object testField1579 = new Integer(1579);
    public Object testField1580 = new Integer(1580);
    public Object testField1581 = new Integer(1581);
    public Object testField1582 = new Integer(1582);
    public Object testField1583 = new Integer(1583);
    public Object testField1584 = new Integer(1584);
    public Object testField1585 = new Integer(1585);
    public Object testField1586 = new Integer(1586);
    public Object testField1587 = new Integer(1587);
    public Object testField1588 = new Integer(1588);
    public Object testField1589 = new Integer(1589);
    public Object testField1590 = new Integer(1590);
    public Object testField1591 = new Integer(1591);
    public Object testField1592 = new Integer(1592);
    public Object testField1593 = new Integer(1593);
    public Object testField1594 = new Integer(1594);
    public Object testField1595 = new Integer(1595);
    public Object testField1596 = new Integer(1596);
    public Object testField1597 = new Integer(1597);
    public Object testField1598 = new Integer(1598);
    public Object testField1599 = new Integer(1599);
    public Object testField1600 = new Integer(1600);
    public Object testField1601 = new Integer(1601);
    public Object testField1602 = new Integer(1602);
    public Object testField1603 = new Integer(1603);
    public Object testField1604 = new Integer(1604);
    public Object testField1605 = new Integer(1605);
    public Object testField1606 = new Integer(1606);
    public Object testField1607 = new Integer(1607);
    public Object testField1608 = new Integer(1608);
    public Object testField1609 = new Integer(1609);
    public Object testField1610 = new Integer(1610);
    public Object testField1611 = new Integer(1611);
    public Object testField1612 = new Integer(1612);
    public Object testField1613 = new Integer(1613);
    public Object testField1614 = new Integer(1614);
    public Object testField1615 = new Integer(1615);
    public Object testField1616 = new Integer(1616);
    public Object testField1617 = new Integer(1617);
    public Object testField1618 = new Integer(1618);
    public Object testField1619 = new Integer(1619);
    public Object testField1620 = new Integer(1620);
    public Object testField1621 = new Integer(1621);
    public Object testField1622 = new Integer(1622);
    public Object testField1623 = new Integer(1623);
    public Object testField1624 = new Integer(1624);
    public Object testField1625 = new Integer(1625);
    public Object testField1626 = new Integer(1626);
    public Object testField1627 = new Integer(1627);
    public Object testField1628 = new Integer(1628);
    public Object testField1629 = new Integer(1629);
    public Object testField1630 = new Integer(1630);
    public Object testField1631 = new Integer(1631);
    public Object testField1632 = new Integer(1632);
    public Object testField1633 = new Integer(1633);
    public Object testField1634 = new Integer(1634);
    public Object testField1635 = new Integer(1635);
    public Object testField1636 = new Integer(1636);
    public Object testField1637 = new Integer(1637);
    public Object testField1638 = new Integer(1638);
    public Object testField1639 = new Integer(1639);
    public Object testField1640 = new Integer(1640);
    public Object testField1641 = new Integer(1641);
    public Object testField1642 = new Integer(1642);
    public Object testField1643 = new Integer(1643);
    public Object testField1644 = new Integer(1644);
    public Object testField1645 = new Integer(1645);
    public Object testField1646 = new Integer(1646);
    public Object testField1647 = new Integer(1647);
    public Object testField1648 = new Integer(1648);
    public Object testField1649 = new Integer(1649);
    public Object testField1650 = new Integer(1650);
    public Object testField1651 = new Integer(1651);
    public Object testField1652 = new Integer(1652);
    public Object testField1653 = new Integer(1653);
    public Object testField1654 = new Integer(1654);
    public Object testField1655 = new Integer(1655);
    public Object testField1656 = new Integer(1656);
    public Object testField1657 = new Integer(1657);
    public Object testField1658 = new Integer(1658);
    public Object testField1659 = new Integer(1659);
    public Object testField1660 = new Integer(1660);
    public Object testField1661 = new Integer(1661);
    public Object testField1662 = new Integer(1662);
    public Object testField1663 = new Integer(1663);
    public Object testField1664 = new Integer(1664);
    public Object testField1665 = new Integer(1665);
    public Object testField1666 = new Integer(1666);
    public Object testField1667 = new Integer(1667);
    public Object testField1668 = new Integer(1668);
    public Object testField1669 = new Integer(1669);
    public Object testField1670 = new Integer(1670);
    public Object testField1671 = new Integer(1671);
    public Object testField1672 = new Integer(1672);
    public Object testField1673 = new Integer(1673);
    public Object testField1674 = new Integer(1674);
    public Object testField1675 = new Integer(1675);
    public Object testField1676 = new Integer(1676);
    public Object testField1677 = new Integer(1677);
    public Object testField1678 = new Integer(1678);
    public Object testField1679 = new Integer(1679);
    public Object testField1680 = new Integer(1680);
    public Object testField1681 = new Integer(1681);
    public Object testField1682 = new Integer(1682);
    public Object testField1683 = new Integer(1683);
    public Object testField1684 = new Integer(1684);
    public Object testField1685 = new Integer(1685);
    public Object testField1686 = new Integer(1686);
    public Object testField1687 = new Integer(1687);
    public Object testField1688 = new Integer(1688);
    public Object testField1689 = new Integer(1689);
    public Object testField1690 = new Integer(1690);
    public Object testField1691 = new Integer(1691);
    public Object testField1692 = new Integer(1692);
    public Object testField1693 = new Integer(1693);
    public Object testField1694 = new Integer(1694);
    public Object testField1695 = new Integer(1695);
    public Object testField1696 = new Integer(1696);
    public Object testField1697 = new Integer(1697);
    public Object testField1698 = new Integer(1698);
    public Object testField1699 = new Integer(1699);
    public Object testField1700 = new Integer(1700);
    public Object testField1701 = new Integer(1701);
    public Object testField1702 = new Integer(1702);
    public Object testField1703 = new Integer(1703);
    public Object testField1704 = new Integer(1704);
    public Object testField1705 = new Integer(1705);
    public Object testField1706 = new Integer(1706);
    public Object testField1707 = new Integer(1707);
    public Object testField1708 = new Integer(1708);
    public Object testField1709 = new Integer(1709);
    public Object testField1710 = new Integer(1710);
    public Object testField1711 = new Integer(1711);
    public Object testField1712 = new Integer(1712);
    public Object testField1713 = new Integer(1713);
    public Object testField1714 = new Integer(1714);
    public Object testField1715 = new Integer(1715);
    public Object testField1716 = new Integer(1716);
    public Object testField1717 = new Integer(1717);
    public Object testField1718 = new Integer(1718);
    public Object testField1719 = new Integer(1719);
    public Object testField1720 = new Integer(1720);
    public Object testField1721 = new Integer(1721);
    public Object testField1722 = new Integer(1722);
    public Object testField1723 = new Integer(1723);
    public Object testField1724 = new Integer(1724);
    public Object testField1725 = new Integer(1725);
    public Object testField1726 = new Integer(1726);
    public Object testField1727 = new Integer(1727);
    public Object testField1728 = new Integer(1728);
    public Object testField1729 = new Integer(1729);
    public Object testField1730 = new Integer(1730);
    public Object testField1731 = new Integer(1731);
    public Object testField1732 = new Integer(1732);
    public Object testField1733 = new Integer(1733);
    public Object testField1734 = new Integer(1734);
    public Object testField1735 = new Integer(1735);
    public Object testField1736 = new Integer(1736);
    public Object testField1737 = new Integer(1737);
    public Object testField1738 = new Integer(1738);
    public Object testField1739 = new Integer(1739);
    public Object testField1740 = new Integer(1740);
    public Object testField1741 = new Integer(1741);
    public Object testField1742 = new Integer(1742);
    public Object testField1743 = new Integer(1743);
    public Object testField1744 = new Integer(1744);
    public Object testField1745 = new Integer(1745);
    public Object testField1746 = new Integer(1746);
    public Object testField1747 = new Integer(1747);
    public Object testField1748 = new Integer(1748);
    public Object testField1749 = new Integer(1749);
    public Object testField1750 = new Integer(1750);
    public Object testField1751 = new Integer(1751);
    public Object testField1752 = new Integer(1752);
    public Object testField1753 = new Integer(1753);
    public Object testField1754 = new Integer(1754);
    public Object testField1755 = new Integer(1755);
    public Object testField1756 = new Integer(1756);
    public Object testField1757 = new Integer(1757);
    public Object testField1758 = new Integer(1758);
    public Object testField1759 = new Integer(1759);
    public Object testField1760 = new Integer(1760);
    public Object testField1761 = new Integer(1761);
    public Object testField1762 = new Integer(1762);
    public Object testField1763 = new Integer(1763);
    public Object testField1764 = new Integer(1764);
    public Object testField1765 = new Integer(1765);
    public Object testField1766 = new Integer(1766);
    public Object testField1767 = new Integer(1767);
    public Object testField1768 = new Integer(1768);
    public Object testField1769 = new Integer(1769);
    public Object testField1770 = new Integer(1770);
    public Object testField1771 = new Integer(1771);
    public Object testField1772 = new Integer(1772);
    public Object testField1773 = new Integer(1773);
    public Object testField1774 = new Integer(1774);
    public Object testField1775 = new Integer(1775);
    public Object testField1776 = new Integer(1776);
    public Object testField1777 = new Integer(1777);
    public Object testField1778 = new Integer(1778);
    public Object testField1779 = new Integer(1779);
    public Object testField1780 = new Integer(1780);
    public Object testField1781 = new Integer(1781);
    public Object testField1782 = new Integer(1782);
    public Object testField1783 = new Integer(1783);
    public Object testField1784 = new Integer(1784);
    public Object testField1785 = new Integer(1785);
    public Object testField1786 = new Integer(1786);
    public Object testField1787 = new Integer(1787);
    public Object testField1788 = new Integer(1788);
    public Object testField1789 = new Integer(1789);
    public Object testField1790 = new Integer(1790);
    public Object testField1791 = new Integer(1791);
    public Object testField1792 = new Integer(1792);
    public Object testField1793 = new Integer(1793);
    public Object testField1794 = new Integer(1794);
    public Object testField1795 = new Integer(1795);
    public Object testField1796 = new Integer(1796);
    public Object testField1797 = new Integer(1797);
    public Object testField1798 = new Integer(1798);
    public Object testField1799 = new Integer(1799);
    public Object testField1800 = new Integer(1800);
    public Object testField1801 = new Integer(1801);
    public Object testField1802 = new Integer(1802);
    public Object testField1803 = new Integer(1803);
    public Object testField1804 = new Integer(1804);
    public Object testField1805 = new Integer(1805);
    public Object testField1806 = new Integer(1806);
    public Object testField1807 = new Integer(1807);
    public Object testField1808 = new Integer(1808);
    public Object testField1809 = new Integer(1809);
    public Object testField1810 = new Integer(1810);
    public Object testField1811 = new Integer(1811);
    public Object testField1812 = new Integer(1812);
    public Object testField1813 = new Integer(1813);
    public Object testField1814 = new Integer(1814);
    public Object testField1815 = new Integer(1815);
    public Object testField1816 = new Integer(1816);
    public Object testField1817 = new Integer(1817);
    public Object testField1818 = new Integer(1818);
    public Object testField1819 = new Integer(1819);
    public Object testField1820 = new Integer(1820);
    public Object testField1821 = new Integer(1821);
    public Object testField1822 = new Integer(1822);
    public Object testField1823 = new Integer(1823);
    public Object testField1824 = new Integer(1824);
    public Object testField1825 = new Integer(1825);
    public Object testField1826 = new Integer(1826);
    public Object testField1827 = new Integer(1827);
    public Object testField1828 = new Integer(1828);
    public Object testField1829 = new Integer(1829);
    public Object testField1830 = new Integer(1830);
    public Object testField1831 = new Integer(1831);
    public Object testField1832 = new Integer(1832);
    public Object testField1833 = new Integer(1833);
    public Object testField1834 = new Integer(1834);
    public Object testField1835 = new Integer(1835);
    public Object testField1836 = new Integer(1836);
    public Object testField1837 = new Integer(1837);
    public Object testField1838 = new Integer(1838);
    public Object testField1839 = new Integer(1839);
    public Object testField1840 = new Integer(1840);
    public Object testField1841 = new Integer(1841);
    public Object testField1842 = new Integer(1842);
    public Object testField1843 = new Integer(1843);
    public Object testField1844 = new Integer(1844);
    public Object testField1845 = new Integer(1845);
    public Object testField1846 = new Integer(1846);
    public Object testField1847 = new Integer(1847);
    public Object testField1848 = new Integer(1848);
    public Object testField1849 = new Integer(1849);
    public Object testField1850 = new Integer(1850);
    public Object testField1851 = new Integer(1851);
    public Object testField1852 = new Integer(1852);
    public Object testField1853 = new Integer(1853);
    public Object testField1854 = new Integer(1854);
    public Object testField1855 = new Integer(1855);
    public Object testField1856 = new Integer(1856);
    public Object testField1857 = new Integer(1857);
    public Object testField1858 = new Integer(1858);
    public Object testField1859 = new Integer(1859);
    public Object testField1860 = new Integer(1860);
    public Object testField1861 = new Integer(1861);
    public Object testField1862 = new Integer(1862);
    public Object testField1863 = new Integer(1863);
    public Object testField1864 = new Integer(1864);
    public Object testField1865 = new Integer(1865);
    public Object testField1866 = new Integer(1866);
    public Object testField1867 = new Integer(1867);
    public Object testField1868 = new Integer(1868);
    public Object testField1869 = new Integer(1869);
    public Object testField1870 = new Integer(1870);
    public Object testField1871 = new Integer(1871);
    public Object testField1872 = new Integer(1872);
    public Object testField1873 = new Integer(1873);
    public Object testField1874 = new Integer(1874);
    public Object testField1875 = new Integer(1875);
    public Object testField1876 = new Integer(1876);
    public Object testField1877 = new Integer(1877);
    public Object testField1878 = new Integer(1878);
    public Object testField1879 = new Integer(1879);
    public Object testField1880 = new Integer(1880);
    public Object testField1881 = new Integer(1881);
    public Object testField1882 = new Integer(1882);
    public Object testField1883 = new Integer(1883);
    public Object testField1884 = new Integer(1884);
    public Object testField1885 = new Integer(1885);
    public Object testField1886 = new Integer(1886);
    public Object testField1887 = new Integer(1887);
    public Object testField1888 = new Integer(1888);
    public Object testField1889 = new Integer(1889);
    public Object testField1890 = new Integer(1890);
    public Object testField1891 = new Integer(1891);
    public Object testField1892 = new Integer(1892);
    public Object testField1893 = new Integer(1893);
    public Object testField1894 = new Integer(1894);
    public Object testField1895 = new Integer(1895);
    public Object testField1896 = new Integer(1896);
    public Object testField1897 = new Integer(1897);
    public Object testField1898 = new Integer(1898);
    public Object testField1899 = new Integer(1899);
    public Object testField1900 = new Integer(1900);
    public Object testField1901 = new Integer(1901);
    public Object testField1902 = new Integer(1902);
    public Object testField1903 = new Integer(1903);
    public Object testField1904 = new Integer(1904);
    public Object testField1905 = new Integer(1905);
    public Object testField1906 = new Integer(1906);
    public Object testField1907 = new Integer(1907);
    public Object testField1908 = new Integer(1908);
    public Object testField1909 = new Integer(1909);
    public Object testField1910 = new Integer(1910);
    public Object testField1911 = new Integer(1911);
    public Object testField1912 = new Integer(1912);
    public Object testField1913 = new Integer(1913);
    public Object testField1914 = new Integer(1914);
    public Object testField1915 = new Integer(1915);
    public Object testField1916 = new Integer(1916);
    public Object testField1917 = new Integer(1917);
    public Object testField1918 = new Integer(1918);
    public Object testField1919 = new Integer(1919);
    public Object testField1920 = new Integer(1920);
    public Object testField1921 = new Integer(1921);
    public Object testField1922 = new Integer(1922);
    public Object testField1923 = new Integer(1923);
    public Object testField1924 = new Integer(1924);
    public Object testField1925 = new Integer(1925);
    public Object testField1926 = new Integer(1926);
    public Object testField1927 = new Integer(1927);
    public Object testField1928 = new Integer(1928);
    public Object testField1929 = new Integer(1929);
    public Object testField1930 = new Integer(1930);
    public Object testField1931 = new Integer(1931);
    public Object testField1932 = new Integer(1932);
    public Object testField1933 = new Integer(1933);
    public Object testField1934 = new Integer(1934);
    public Object testField1935 = new Integer(1935);
    public Object testField1936 = new Integer(1936);
    public Object testField1937 = new Integer(1937);
    public Object testField1938 = new Integer(1938);
    public Object testField1939 = new Integer(1939);
    public Object testField1940 = new Integer(1940);
    public Object testField1941 = new Integer(1941);
    public Object testField1942 = new Integer(1942);
    public Object testField1943 = new Integer(1943);
    public Object testField1944 = new Integer(1944);
    public Object testField1945 = new Integer(1945);
    public Object testField1946 = new Integer(1946);
    public Object testField1947 = new Integer(1947);
    public Object testField1948 = new Integer(1948);
    public Object testField1949 = new Integer(1949);
    public Object testField1950 = new Integer(1950);
    public Object testField1951 = new Integer(1951);
    public Object testField1952 = new Integer(1952);
    public Object testField1953 = new Integer(1953);
    public Object testField1954 = new Integer(1954);
    public Object testField1955 = new Integer(1955);
    public Object testField1956 = new Integer(1956);
    public Object testField1957 = new Integer(1957);
    public Object testField1958 = new Integer(1958);
    public Object testField1959 = new Integer(1959);
    public Object testField1960 = new Integer(1960);
    public Object testField1961 = new Integer(1961);
    public Object testField1962 = new Integer(1962);
    public Object testField1963 = new Integer(1963);
    public Object testField1964 = new Integer(1964);
    public Object testField1965 = new Integer(1965);
    public Object testField1966 = new Integer(1966);
    public Object testField1967 = new Integer(1967);
    public Object testField1968 = new Integer(1968);
    public Object testField1969 = new Integer(1969);
    public Object testField1970 = new Integer(1970);
    public Object testField1971 = new Integer(1971);
    public Object testField1972 = new Integer(1972);
    public Object testField1973 = new Integer(1973);
    public Object testField1974 = new Integer(1974);
    public Object testField1975 = new Integer(1975);
    public Object testField1976 = new Integer(1976);
    public Object testField1977 = new Integer(1977);
    public Object testField1978 = new Integer(1978);
    public Object testField1979 = new Integer(1979);
    public Object testField1980 = new Integer(1980);
    public Object testField1981 = new Integer(1981);
    public Object testField1982 = new Integer(1982);
    public Object testField1983 = new Integer(1983);
    public Object testField1984 = new Integer(1984);
    public Object testField1985 = new Integer(1985);
    public Object testField1986 = new Integer(1986);
    public Object testField1987 = new Integer(1987);
    public Object testField1988 = new Integer(1988);
    public Object testField1989 = new Integer(1989);
    public Object testField1990 = new Integer(1990);
    public Object testField1991 = new Integer(1991);
    public Object testField1992 = new Integer(1992);
    public Object testField1993 = new Integer(1993);
    public Object testField1994 = new Integer(1994);
    public Object testField1995 = new Integer(1995);
    public Object testField1996 = new Integer(1996);
    public Object testField1997 = new Integer(1997);
    public Object testField1998 = new Integer(1998);
    public Object testField1999 = new Integer(1999);
}
