/*
 * Copyright (C) 2008 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import java.lang.reflect.Field;

/**
 * Try some stuff with enumerations.
 */
public class Main {
    public enum Shubbery { GROUND, CRAWLING, HANGING }

    public static void main(String[] args) {
        Field field;
        try {
            field = Shubbery.class.getDeclaredField("CRAWLING");
        } catch (NoSuchFieldException nsfe) {
            throw new RuntimeException(nsfe);
        }

        System.out.println("found field " + field.getName());
        System.out.println("  synthetic? " + field.isSynthetic());
        System.out.println("  enum? " + field.isEnumConstant());
    }
}
