/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#if ASM_DEFINE_INCLUDE_DEPENDENCIES
#include "mirror/dex_cache.h"
#endif

ASM_DEFINE(METHOD_DEX_CACHE_SIZE_MINUS_ONE,
           art::mirror::DexCache::kDexCacheMethodCacheSize - 1)
ASM_DEFINE(MIRROR_DEX_CACHE_RESOLVED_METHODS_OFFSET,
           art::mirror::DexCache::ResolvedMethodsOffset().Int32Value())
ASM_DEFINE(STRING_DEX_CACHE_ELEMENT_SIZE,
           sizeof(art::mirror::StringDexCachePair))
ASM_DEFINE(STRING_DEX_CACHE_ELEMENT_SIZE_SHIFT,
           art::WhichPowerOf2(sizeof(art::mirror::StringDexCachePair)))
ASM_DEFINE(STRING_DEX_CACHE_HASH_BITS,
           art::LeastSignificantBit(art::mirror::DexCache::kDexCacheStringCacheSize))
ASM_DEFINE(STRING_DEX_CACHE_SIZE_MINUS_ONE,
           art::mirror::DexCache::kDexCacheStringCacheSize - 1)
ASM_DEFINE(METHOD_DEX_CACHE_HASH_BITS,
           art::LeastSignificantBit(art::mirror::DexCache::kDexCacheMethodCacheSize))
