/*
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
  public static void main(String[] args) {
    System.loadLibrary(args[0]);
    holdFpTemporaries();
    System.out.println("Done");
  }

  public static void caller(int a, int b, long c) {
    $noinline$callee(a, b, c);
  }

  // This method is "no inline", in order to generate the
  // bad floating point use at the call site.
  public static void $noinline$callee(int a, int b, long c) {
  }

  public native static void holdFpTemporaries();
}
