/*
 * Copyright (C) 2014 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "bit_field.h"
#include "globals.h"

#include "gtest/gtest.h"

namespace art {

TEST(BitFields, Test1) {
  class TestBitFields : public BitField<int32_t, 1, 8> {};
  ASSERT_TRUE(TestBitFields::IsValid(16));
  ASSERT_TRUE(!TestBitFields::IsValid(256));
  ASSERT_EQ(0x00ffU, TestBitFields::Mask());
  ASSERT_EQ(0x001feU, TestBitFields::MaskInPlace());
  ASSERT_EQ(1, TestBitFields::Shift());
  ASSERT_EQ(8, TestBitFields::BitSize());
  ASSERT_EQ(32U, TestBitFields::Encode(16));
  ASSERT_EQ(16, TestBitFields::Decode(32));
  ASSERT_EQ(2U, TestBitFields::Update(1, 16));
}

}  // namespace art
