/*
 * Copyright (C) 2006 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import otherpackage.PublicAccess;

public class Main {
    public static void main(String[] args) {
        System.out.println("access test");

        PublicAccess pa = new PublicAccess();
        pa.main();

        System.out.println(Iface.X);
        System.out.println(Iface2.Y);
        System.out.println(Iface2.A);
        System.out.println(Iface2.B);
    }
}
