/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>

#include "memfd.h"

TEST(memfd, basic) {
  errno = 0;
  int fd = art::memfd_create("memfd_create_test", 0);
  if (fd < 0) {
    ASSERT_EQ(ENOSYS, errno);
    GTEST_LOG_(INFO) << "memfd_create not supported, skipping test.";
    return;
  }
  ASSERT_TRUE(close(fd) == 0 || errno != EBADF);
}
