/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef ART_SIGCHAINLIB_LOG_H_
#define ART_SIGCHAINLIB_LOG_H_

#if defined(__ANDROID__)

#include <async_safe/log.h>

#define log(...) async_safe_format_log(ANDROID_LOG_ERROR, "libsigchain", __VA_ARGS__)
#define fatal async_safe_fatal

#else

#include <stdarg.h>
#include <stdio.h>

static void log(const char* format, ...) {
  va_list ap;
  va_start(ap, format);
  vprintf(format, ap);
  va_end(ap);

  printf("\n");
}

#define fatal(...) log(__VA_ARGS__); abort()

#endif

#endif  // ART_SIGCHAINLIB_LOG_H_
