/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


public class Main {
    public static void main(String[] args) throws Throwable {
        try {
            // Class BadField is defined in BadField.smali.
            Class<?> c = Class.forName("BadField");
            System.out.println("Not reached");
            c.newInstance();
        } catch (NoClassDefFoundError expected) {
            // The NoClassDefFoundError is for the field widget in class BadField.
            if (expected.getMessage().equals("Failed resolution of: LWidget;")) {
                System.out.println("passed");
            } else {
                System.out.println("failed: " + expected.getMessage());
            }
        }
    }
}
