/*
 * Copyright (C) 2019 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
  public static boolean alwaysThrows() {
    throw new Error("");
  }

  public static void test() {
    alwaysThrows();
    while (condition) {
      int a = 2;
      while (otherCondition) {
        a = 3;
      }
      staticField = a;
    }
  }

  public static void main(String[] args) throws Exception {
    try {
      test();
      throw new Exception("Unexpected exception");
    } catch (Error e) {
      // Expected.
    }
  }

  static boolean condition = false;
  static boolean otherCondition = false;
  static int staticField = 1;
}
