/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#if !defined(ASM_DEFINE_INCLUDE_DEPENDENCIES)
#define ASM_DEFINE_INCLUDE_DEPENDENCIES 1
#endif

#include "globals.def"
#include "art_field.def"
#include "art_method.def"
#include "code_item.def"
#include "lockword.def"
#include "mirror_array.def"
#include "mirror_class.def"
#include "mirror_dex_cache.def"
#include "mirror_object.def"
#include "mirror_string.def"
#include "osr.def"
#include "profiling_info.def"
#include "rosalloc.def"
#include "runtime.def"
#include "shadow_frame.def"
#include "thread.def"
