/*
 * Copyright (C) 2020 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
  public static void sayHi(String klass) throws Exception {
    try {
      System.out.println(klass);
      IFace iface = (IFace)Class.forName(klass).newInstance();
      iface.sayHi();
    } catch (Exception e) {
      System.out.println("Exception thrown!");
      System.out.println(e);
    }
  }
  public static void main(String[] args) throws Exception {
    sayHi("Concrete1");
    sayHi("Concrete2");
    sayHi("Concrete3");
  }
}
