/*
 * Copyright (C) 2016 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Main {
  Main() {
    // Will be quickened with RETURN_VOID_NO_BARRIER.
  }

  public static void main(String[] args) {
    Main m = new Main();
    Object o = m;
    // The call and field accesses will be quickened.
    m.foo(m.a);

    // The checkcast will be quickened.
    m.foo(((Main)o).a);
  }

  int a;
  void foo(int a) {
    System.out.println("In foo");
  }
}

