/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "jni.h"

namespace art {
namespace Test1960ObsoleteJitMultithreaded {

extern "C" JNIEXPORT void JNICALL Java_Transform_nativeSayHi(JNIEnv* env,
                                                             jobject thiz,
                                                             jobject arg1,
                                                             jobject arg2) {
  env->CallVoidMethod(
      thiz,
      env->GetMethodID(env->GetObjectClass(thiz),
                       "sayHi",
                       "(Ljava/util/function/Consumer;Ljava/util/function/Consumer;)V"),
      arg1,
      arg2);
}

}  // namespace Test1960ObsoleteJitMultithreaded
}  // namespace art
