/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package art.constmethodhandle;

// We call one of these using a constant method-handle. Which one is called is changed by the
// redefinition.
public class Responses {
  public static void sayHi() {
    System.out.println("Hello");
  }

  public static void sayBye() {
    System.out.println("Goodbye");
  }
}
