/*
 * Copyright (C) 2018 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#if ASM_DEFINE_INCLUDE_DEPENDENCIES
#include "gc/allocator/rosalloc.h"
#endif

ASM_DEFINE(ROSALLOC_BRACKET_QUANTUM_SIZE_MASK,
           art::gc::allocator::RosAlloc::kThreadLocalBracketQuantumSize - 1)
ASM_DEFINE(ROSALLOC_BRACKET_QUANTUM_SIZE_MASK_TOGGLED32,
           ~static_cast<uint32_t>(art::gc::allocator::RosAlloc::kThreadLocalBracketQuantumSize - 1))
ASM_DEFINE(ROSALLOC_BRACKET_QUANTUM_SIZE_MASK_TOGGLED64,
           ~static_cast<uint64_t>(art::gc::allocator::RosAlloc::kThreadLocalBracketQuantumSize - 1))
ASM_DEFINE(ROSALLOC_BRACKET_QUANTUM_SIZE_SHIFT,
           art::gc::allocator::RosAlloc::kThreadLocalBracketQuantumSizeShift)
ASM_DEFINE(ROSALLOC_MAX_THREAD_LOCAL_BRACKET_SIZE,
           art::gc::allocator::RosAlloc::kMaxThreadLocalBracketSize)
ASM_DEFINE(ROSALLOC_RUN_FREE_LIST_HEAD_OFFSET,
           art::gc::allocator::RosAlloc::RunFreeListHeadOffset())
ASM_DEFINE(ROSALLOC_RUN_FREE_LIST_OFFSET,
           art::gc::allocator::RosAlloc::RunFreeListOffset())
ASM_DEFINE(ROSALLOC_RUN_FREE_LIST_SIZE_OFFSET,
           art::gc::allocator::RosAlloc::RunFreeListSizeOffset())
ASM_DEFINE(ROSALLOC_SLOT_NEXT_OFFSET,
           art::gc::allocator::RosAlloc::RunSlotNextOffset())
