/*
 * Copyright (C) 2014 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class StaticArrayFieldsTest {
    public static boolean[] booleanArrayField = new boolean[1];
    public static byte[]    byteArrayField = new byte[1];
    public static char []   charArrayField = new char[1];
    public static short[]   shortArrayField = new short[1];
    public static int[]     intArrayField = new int[1];
    public static long[]    longArrayField = new long[1];
    public static float[]   floatArrayField = new float[1];
    public static double[]  doubleArrayField = new double[1];
    public static Object[]  objectArrayField = new Object[1];
}
