/*
 * Copyright (C) 2014 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

public class Second implements Inf1, Inf2, Inf3, Inf4, Inf5, Inf6, Inf7, Inf8 {
  public void zcall() {
    System.out.println("Second::zcall");
  }

  public void zcall1() {
    System.out.println("Second::zcall1");
  }

  public void zcall2() {
    System.out.println("Second::zcall2");
  }

  public void zcall3() {
    System.out.println("Second::zcall3");
  }

  public void zcall4() {
    System.out.println("Second::zcall4");
  }

  public void zcall5() {
    System.out.println("Second::zcall5");
  }

  public void zcall6() {
    System.out.println("Second::zcall6");
  }

  public void zcall7() {
    System.out.println("Second::zcall7");
  }

  public void zcall8() {
    System.out.println("Second::zcall8");
  }

  public void zcall9() {
    System.out.println("Second::zcall9");
  }
}
